#include "BasicAIModule.h"
#include "log.h"
#include <windows.h>
#include "UnitGroup.h"

using namespace BWAPI;

DWORD WINAPI AnalyzeThread(bool & analyzed);

void BasicAIModule::onStart()
{
	/* Run once 
	/*
	scai::CBRModule cbr3("bwapi-data\\caseLib\\small.index");
	cbr3.create_composite_database("bwapi-data\\caseLib\\small.clustered.index");
	
	scai::CBRModule cbr2("bwapi-data\\caseLib\\medium.index");
	cbr2.create_composite_database("bwapi-data\\caseLib\\medium.clustered.index");
	
	scai::CBRModule cbr1("bwapi-data\\caseLib\\default.index");
	cbr1.create_composite_database("bwapi-data\\caseLib\\default.clustered.index");
	BWAPI::Broodwar->leaveGame();
	scai::log("Ran one-offs!");
	return;*/
	/* end run once */
	scai::log("-------------- NEW GAME ---------------");

	scai::log("Running tests");
	tester.runTests();
	scai::log("....ran tests");

	//BWAPI::Broodwar->setLocalSpeed(0);


	if (Broodwar->isReplay()) {
		//Broodwar->setGUI(false);
		//Broodwar->setLocalSpeed(0);
		return;
	}
	// Enable some cheat flags
	Broodwar->enableFlag(Flag::UserInput);
	//Broodwar->enableFlag(Flag::CompleteMapInformation);

	BWTA::readMap();
	Broodwar->printf("BEGIN: Analyzing map... this may take a minute");
	CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)AnalyzeThread, &analyzed, 0, NULL);
	//BWTA::analyze();
	

	BWAPI::Race race = Broodwar->self()->getRace();
	BWAPI::Race enemyRace = Broodwar->enemy()->getRace();
	BWAPI::UnitType workerType=race.getWorker();


	if (is_testrun) {
		test_runner->start(BWAPI::Broodwar->mapName());
		planner = new scai::Planner(comm, test_runner->get_db(), test_runner);
	} else {
		std::string default_database = "bwapi-data\\caseLib\\default.index";
		scai::log("Creating planner with default database");
		planner = new scai::Planner(comm, default_database, test_runner);
		scai::log("Created planner");
	}
	executive = new scai::Executive(comm);
	reactive = new scai::Reactive(comm);

	comm.addLayer(planner);
	comm.addLayer(executive);
	comm.addLayer(reactive);

	executive->onStart();
	BWAPI::Broodwar->setGUI(settings.gui);
	BWAPI::Broodwar->setLocalSpeed(settings.local_speed);
}

BasicAIModule::~BasicAIModule()
{
	delete test_runner;
}

void BasicAIModule::onEnd(bool isWinner)
{
	scai::log("onEnd(%d)",isWinner);
	if (is_testrun) {
		test_runner->end((isWinner ? scai::TestRunner::Win : scai::TestRunner::Loss), BWAPI::Broodwar->enemy()->getRace().getName());
		bool check = test_runner->check_stop();
		scai::log("Testrunner did run with outcome %d", check);
	}
}
void BasicAIModule::onFrame()
{
	if (Broodwar->isReplay()) return;
	int frame = BWAPI::Broodwar->getFrameCount();
	try {
		//scai::log("pre reactive");
		reactive->onFrame();
		//scai::log("pre exec");
		executive->onFrame();
		//scai::log("pre plan");
		planner->onFrame();
		//scai::log("frame end");
		
		//TRACE(reactive->onFrame(), "reactive");
		
		//TRACE(executive->onFrame(), "executive");
		
		//TRACE(planner->onFrame(), "planner");
		
	} catch (std::exception e) {
		scai::log ("ERROR: %s in frame %d", e.what(), frame);
		Broodwar->printf("ERROR: %s", e.what());

		if (is_testrun) {
			test_runner->crash(e.what());
			BWAPI::Broodwar->leaveGame();
		}

	} catch (...) {
		scai::log ("ERROR: Unknown exception in frame %d", frame);
		Broodwar->printf("ERROR: Unknown exception");
	}

	/*
	scai::UnitGroup all = scai::Global::i()->all().mine();
	for (int i = 0; i != all.size(); ++i)
		Broodwar->drawTextScreen(5, 5 + 10*i, "%d: %s", i, all.at(i)->getType().getName().c_str());
	*/
}

void BasicAIModule::onUnitCreate(BWAPI::Unit* unit) {
	if (Broodwar->isReplay()) return;

	int frame = BWAPI::Broodwar->getFrameCount();
	unit->setClientInfo(new scai::UnitData(frame));
	executive->onUnitCreate(unit);
}

void BasicAIModule::onUnitMorph(BWAPI::Unit* unit) {
	if (Broodwar->isReplay()) return;

	if (unit->getType() == BWAPI::UnitTypes::Protoss_Assimilator) {
		scai::log ("onunitmorph assimilator");
		int frame = BWAPI::Broodwar->getFrameCount();
		unit->setClientInfo(new scai::UnitData(frame));
		executive->onUnitCreate(unit);
	}
}

void BasicAIModule::onUnitDestroy(BWAPI::Unit* unit)
{
	if (Broodwar->isReplay()) return;

	delete unit->getClientInfo();
}

void BasicAIModule::onUnitDiscover(BWAPI::Unit* unit)
{
	if (Broodwar->isReplay()) return;
}
void BasicAIModule::onUnitEvade(BWAPI::Unit* unit)
{
	if (Broodwar->isReplay()) return;
}

void BasicAIModule::onUnitRenegade(BWAPI::Unit* unit)
{
	if (Broodwar->isReplay()) return;
}

void BasicAIModule::onSendText(std::string text)
{	
	Broodwar->sendText("%s",text.c_str());
}



DWORD WINAPI AnalyzeThread(bool & analyzed)
{
	BWTA::analyze();
	Broodwar->printf("END:   Analyzing map... this may take a minute");
	analyzed = true;
	return 0;
}
