#pragma once
#include <BWAPI.h>
#include <BWTA.h>

#include "ILayer.h"
#include "Planner.h"
#include "Executive.h"
#include "Reactive.h"
#include "Global.h"
#include "log.h"



#include "Communication.h"
#include "PlanItem.h"

#include "TestClass.h"
#include "TestRunner.h"

namespace scai {
	/**
	 *  Settings determine wether we're running tests, which test file to use,
	 *  and BWAPI settings (gui, speed etc.)
	 */
	struct Settings {
		std::string testfile;
		bool gui;
		int local_speed;

		/* Settings are always stored in settings.txt */
		void load() {
			scai::log ("in settings::load()");
			std::ifstream ifs("settings.txt");
			if (ifs) {
				std::getline(ifs, testfile);
				ifs >> gui; ifs.ignore();
				ifs >> local_speed; ifs.ignore();
			}
		}
	};
}

class BasicAIModule : public BWAPI::AIModule {
public:
	BasicAIModule() {
		settings.load();
		is_testrun = (settings.testfile != "");
		if (is_testrun) {
			test_runner = new scai::TestRunner();
			scai::log ("testrunner::load_settings()");
			test_runner->load_test(settings.testfile);
		} else {
			test_runner = new scai::MockTestRunner();
		}
	}

	scai::ILayer *planner, *executive, *reactive;
	scai::Communication comm;

	virtual void onStart();
	virtual void onEnd(bool isWinner);
	virtual void onFrame();
	virtual void onUnitDiscover(BWAPI::Unit* unit);
	virtual void onUnitEvade(BWAPI::Unit* unit);
	virtual void onUnitMorph(BWAPI::Unit* unit);
	virtual void onUnitRenegade(BWAPI::Unit* unit);
	virtual void onUnitCreate(BWAPI::Unit* unit);
	virtual void onUnitDestroy(BWAPI::Unit* unit);
	virtual void onSendText(std::string text);
	
	~BasicAIModule();
	
	void showStats(); //not part of BWAPI::AIModule
	void showPlayers();
	void showForces();
	
	bool analyzed;
	bool showManagerAssignments;
	std::map<BWAPI::Unit*,BWAPI::UnitType> buildings;

	scai::TestClass tester;
	
	bool is_testrun;
	scai::TestRunner* test_runner;
	scai::Settings settings;
};