#pragma once

#include <list>
#include <BWAPI.h>

namespace scai {
	struct BuildingManager {
		scai::MyUnitGroup under_construction;
		scai::MyUnitGroup buildings;

		static bool isMyBuilding(BWAPI::Unit* unit) {
			return (
				(unit->getPlayer() == BWAPI::Broodwar->self()) &&
				(unit->getType().isBuilding())
			);
		}

		/**
		 *  Get the remaining time until this building is ready to produce more units.
		 */
		int RemainingTime(BWAPI::Unit* unit) {
			int time = unit->getRemainingTrainTime();
			std::list<BWAPI::UnitType> q = unit->getTrainingQueue();
			for (std::list<BWAPI::UnitType>::iterator it = q.begin(); it != q.end(); ++it)
				time += it->buildTime();
			return time;
		}

		void onUnitCreate(BWAPI::Unit* unit) {
			if (isMyBuilding(unit)) {
				
				if (unit->isCompleted()) {
					buildings.add(unit);
				} else {
					under_construction.add(unit);
				}
			}
		}

		MyUnitGroup get_producer(const BWAPI::UnitType& type) {
			MyUnitGroup producers = buildings.producers(type);
			if (producers.isEmpty()) return MyUnitGroup();
			
			size_t b = 0;
			int rem_time = RemainingTime(producers.at(b));
			for (size_t i = 1; i < producers.size(); ++i) {
				int rt = RemainingTime(producers.at(i));
				if (rt < rem_time) {
					b = i;
					rem_time = rt;
				}
			}
			return MyUnitGroup(producers.at(b));
		}

		int RemainingUpgradeTime(BWAPI::Unit* unit) {
			return unit->getRemainingUpgradeTime();
		}

		MyUnitGroup get_upgraders(const BWAPI::UpgradeType& type) {
			MyUnitGroup upgraders = buildings.upgraders(type).is_completed();
			if (upgraders.isEmpty()) return upgraders;

			size_t b = 0;
			int rem_time = RemainingUpgradeTime(upgraders.at(b));
			for (size_t i = 1; i < upgraders.size(); ++i) {
				int rt = RemainingUpgradeTime(upgraders.at(i));
				if (rt < rem_time) {
					b = i;
					rem_time = rt;
				}
			}
			return MyUnitGroup(upgraders.at(b));
		}

		void onFrame() {
			MyUnitGroup completed = under_construction.is_completed();
			if (!completed.isEmpty()) {
				buildings.add(completed);
				under_construction.rem(completed);
			}
		}
	};
}