#pragma once

#include <BWAPI.h>
#include "Case.h"
#include "Gamestate.h"
#include <vector>
#include <map>
#include <algorithm>

#include "log.h"


namespace scai {
	//TODO: Fix weights, and test in general :)
	class CaseMatcher {
		template <typename T> T Min(T a, T b) { return (a < b) ? a : b; }
	public:
		/*To be used for comparing 2 different cases.
		* Does not consider plan or opponent model equality, only gamestate and frames
		*/
		float caseCompare(const Case& item1, const Case& item2, float minMatchValue) {
			float ans = 0.0f;

			float wState = 20.0f;
			float wFrames = 1.0f;

			float wTotal = wState + wFrames;
			ans += wState * gamestateMatch(item1.state, item2.state, minMatchValue);
			ans += wFrames * filter(cInts(item1.frames, item2.frames), minMatchValue);


			return (ans/wTotal);
		}

		//Used by planner when it needs a new case
		//TODO: Better matching when current gamestate has more than matching case
		float plannerCompare(const Case& item, const Gamestate& state, float minMatchValue) {
			return gamestateMatch(item.state, state, minMatchValue);
		}



	private:
		
		//minMatchValue is the minimum value each item must be above for a complete match
		float gamestateMatch(const Gamestate& state1, const Gamestate& state, float minMatchValue) {
			float ans = 0.0f;

			float wResources = 1.0f;
			float wHarvesters = 1.0f;
			float wResearch = 1.0f;
			float wMUnits = 4.0f;
			float wEUnits = 4.0f;
			float wMBuildings = 4.0f;
			float wEBuildings = 4.0f;
			float wBases = 1.0f;

			float wTotal = 
				wResources + 
				wHarvesters + 
				wResearch + 
				wMUnits + 
				wEUnits + 
				wMBuildings + 
				wEBuildings + 
				wBases;

//			Gamestate& caseState = item.state;

			//Assuming map is prechecked, races are equal
			//Find mineral equality
			ans += wResources * filter(mineralsMatch(state1, state), minMatchValue);

			//Find harvester equality
			ans += wHarvesters * filter(harvestersMatch(state1, state), minMatchValue);

			//Find research equality
			ans += wResearch * filter(researchMatch(state1, state), minMatchValue);

			//Find unit equality, TODO: Make similar units match (melee for melee etc), and more detailed
			ans += wMUnits * filter(unitMatch(state1.mUnits, state.mUnits), minMatchValue);

			ans += wEUnits * filter(unitMatch(state1.eUnits, state.eUnits), minMatchValue);

			ans += wMBuildings * filter(buildingMatch(state1.mBuildings, state.mBuildings), minMatchValue);

			ans += wEBuildings * filter(buildingMatch(state1.eBuildings, state.eBuildings), minMatchValue);

			ans += wBases * filter(baseMatch(state1, state), minMatchValue);

			return (ans/wTotal);
		}

		float planMatch(Plan plan1, Plan plan2) {
			float matched = 0.0f;
		}
		
		
		float mineralsMatch(const Gamestate& state1, const Gamestate& state2) {
			return ((cInts(state1.mMinerals, state2.mMinerals) + 
				cInts(state1.mGas, state2.mGas) +
				cFloats(state1.mMineralsRate, state2.mMineralsRate) +
				cFloats(state1.mGasRate, state2.mGasRate))/4.0f);
		}

		float harvestersMatch(const Gamestate& state1, const Gamestate& state2) {
			return ((cInts(state1.mHarvesters, state2.mHarvesters) + 
				cInts(state1.eHarvesters, state2.eHarvesters)) / 2.0f);
		}

		float researchMatch(const Gamestate& state1, const Gamestate& state2) {
			float total = 0.0f;
			float matched = 0.0f;

			//for each research in casestate, see if the current gamestate has an equal or subset of the research, add match times 2 for each state
			for(std::map<int, int>::const_iterator it = state1.mResearch.begin() ; it != state1.mResearch.end() ; ++it) {
				total += (float)it->second;
				//If our gamestate has higher value in research, consider it a 100% match
				int lvl = 0;
				std::map<int, int>::const_iterator it2 = state2.mResearch.find(it->first);
				if (it2 != state2.mResearch.end()) lvl = it2->second;
				if(lvl > it->second) {
					matched += (float)it->second*2.0f;
				} else {
					matched += (float)lvl*2.0f;
				}
			}
			//Add each research we have in state2 for total
			for(std::map<int, int>::const_iterator it = state2.mResearch.begin() ; it != state2.mResearch.end() ; ++it) {
				total += (float)it->second;
			}


			for(std::map<int, int>::const_iterator it = state1.eResearch.begin() ; it != state1.eResearch.end() ; ++it) {
				total += (float)it->second;
				
				//If our gamestate has higher value in research, consider it a 100% match
				int lvl = 0;
				std::map<int, int>::const_iterator it2 = state2.eResearch.find(it->first);
				if (it2 != state2.eResearch.end()) lvl = it2->second;
				if(lvl > it->second) {
					matched += (float)it->second*2.0f;
				} else {
					matched += (float)lvl*2.0f;
				}
			}
			//Add each research we have in state2 for total
			for(std::map<int, int>::const_iterator it = state2.eResearch.begin() ; it != state2.eResearch.end() ; ++it) {
				total += (float)it->second;
			}


			if(total == 0.0f) {
				//to avoid divide by zero errors, no research in either state means we get 100% match
				return 1.0f;
			} else if(matched == 0.0f) { //No matches means we need to add a minimum value to see how far off the matching is
				matched += 1.0f;
				total += 1.0f;
				return (matched/total);
			} else {
				return (matched/total);
			}
		}

		//TODO: Fix so its more detailed
		float unitMatch(const std::vector<Gamestate::Units>& state1, const std::vector<Gamestate::Units>& state2) {
			int state1_totalValue = 0;
			int state2_totalValue = 0;

			float matched = 0.0f;
			float total = 0.0f;
			
			//Find total value in state2
			for(std::vector<Gamestate::Units>::const_iterator it = state2.begin() ; it != state2.end() ; ++it) {
				total += (float)(it->count);
				BWAPI::UnitType temp = BWAPI::UnitType(it->type_id);
				state2_totalValue += temp.mineralPrice() * it->count + temp.gasPrice()*it->count;
			}

			//Find total value in state1
			for(std::vector<Gamestate::Units>::const_iterator it = state1.begin() ; it != state1.end() ; ++it) {
				total += (float)(it->count);
				BWAPI::UnitType temp = BWAPI::UnitType(it->type_id);
				state1_totalValue += temp.mineralPrice()*it->count + temp.gasPrice()*it->count;
			}

			// A map from type_id -> count
			std::map<int, int> units1;
			for (size_t i = 0; i != state1.size(); ++i) {
				std::map<int, int>::iterator it2 = units1.find( state1[i].type_id );
				if (it2 != units1.end()) it2->second += state1[i].count;
				else units1[state1[i].type_id] = state1[i].count;
			}
			std::map<int, int> units2;
			for (size_t i = 0; i != state2.size(); ++i) {
				std::map<int, int>::iterator it2 = units2.find( state2[i].type_id );
				if (it2 != units2.end()) it2->second += state2[i].count;
				else units2[state2[i].type_id] = state2[i].count;
			}
			for (std::map<int, int>::iterator it = units1.begin(); it != units1.end(); ++it) {
				std::map<int, int>::iterator it2 = units2.find( it->first );
				if (it2 != units2.end()) matched += 2.0 * Min(it->second, it2->second);
			}
			return ((cInts(state1_totalValue, state2_totalValue) + cFloats(matched, total))/2.0f);
		}

		float buildingMatch(const std::vector<Gamestate::Units>& state1, const std::vector<Gamestate::Units>& state2) {
			return unitMatch(state1, state2);
			/*
			int state1_totalValue = 0;
			int state2_totalValue = 0;

			float matched = 0.0f;
			float total = 0.0f;
			
			//Find total value in state2
			for(std::vector<Gamestate::Units>::const_iterator it = state2.begin() ; it != state2.end() ; ++it) {
				total += 1.0f;
				BWAPI::UnitType temp = BWAPI::UnitType(it->type_id);
				state2_totalValue += temp.mineralPrice() + temp.gasPrice();
			}

			//Find total value in state1 and see if we have an equivalent building in state 2. If yes, add to matched and remove it from the list
			for(std::vector<Gamestate::Units>::const_iterator it = state1.begin() ; it != state1.end() ; ++it) {
				total += 1.0f;
				BWAPI::UnitType temp = BWAPI::UnitType(it->type_id);
				state1_totalValue += temp.mineralPrice() + temp.gasPrice();

				for(std::vector<Gamestate::Units>::const_iterator it2 = state2.begin() ; it2 != state2.end() ; ++it2) {
					if(it->type_id == it2->type_id) {
						matched += 2.0f;
						state2.erase(it2);
						break;
					}
				}
			}
			return ((cInts(state1_totalValue, state2_totalValue) + cFloats(matched, total))/2.0f);
			*/
		}

		float baseMatch(const Gamestate& state1, const Gamestate& state2) {
			return ((cInts(state1.mBases.size(), state2.mBases.size()) +
				cInts(state1.eBases.size(), state2.eBases.size()))/2.0f);
		}


		//Finds the similarity between two numbers. 
		float cInts(int first, int second) {
			//To make sure we dont have zero of either.
			if(first == 0 || second == 0) {
				first++;
				second++;
			}
			float ans = 0.0f;
			if(first < second) {
				ans = (float)((float)first/(float)second);
			} else {
				ans = (float)((float)second / (float)first);
			}
			return ans;
		}

		float cFloats(float first, float second) {
			//To make sure we dont have zero of either.
			if(first == 0.0f || second == 0.0f) {
				first += 1.0f;
				second += 1.0f;
			}
			//TODO: Make sure the input isnt negative!
			if(first < 0.0f && second < 0.0f) {
				return 1.0f;
			}
			if(first < 0.0f) {
				first = 1.0f;
				second += 1.0f;
			}
			if(second < 0.0f) {
				second = 1.0f;
				first += 1.0f;
			}
			float ans = 0.0f;
			if(first < second) {
				ans = first/second;
			} else {
				ans = second/first;
			}
			return ans;
		}

		float filter(float value, float minMatchValue) {
			if(value >= minMatchValue) {
				return value;
			} else {
				return 0.0f;
			}
		}


	};
}