#pragma once

#include "BasicAIModule.h"
#include "MyUnitGroup.h"
#include <BWAPI.h>
#include <BuildingPlacer.h>
#include "IdFactory.h"
#include "Resource.h"

namespace scai {
	struct ResMan {
		ResMan(): last_min(50), last_gas(0) {}
		static const int frames = 1440;

		int last_min;
		int last_gas;
		/* Pair of frame, delta */
		std::vector<std::pair<int, int> > min;
		std::vector<std::pair<int, int> > gas;

		void onFrame() {
			int frame = BWAPI::Broodwar->getFrameCount();
			int minerals = BWAPI::Broodwar->self()->minerals();
			int d_min = minerals - last_min;
			last_min = minerals;
			if (d_min > 0) {
				min.push_back(std::pair<int, int>(frame, d_min));
			}
			while (min.size() && min.front().first < (frame - frames)) {
				min.erase(min.begin());
			}

			int gaz = BWAPI::Broodwar->self()->gas();
			int d_gas = gaz - last_gas;
			last_gas = gaz;
			if (d_gas > 0) {
				gas.push_back(std::pair<int, int>(frame, d_gas));
			}
			while (gas.size() && gas.front().first < (frame - frames))
				gas.erase(gas.begin());
		}

		float getRate(const std::vector<std::pair<int, int> >& res) {
			int sum = 0;
			for (int i = 0; i != res.size(); ++i)
				sum += res[i].second;
			return ((float) sum) / frames;
		}

		float getMineralRate() { return getRate(min); }
		float getGasRate() { return getRate(gas); }
	};

	

	struct Global {
	private:
		Global(): committedMinerals(0), committedGas(0), base_index(0) {};
		Global(Global const&) {};
		Global& operator = (Global const&) {};
		Gamestate state;
		
		ResMan resman;
		BuildingPlacer placer;
		int committedMinerals;
		int committedGas;
		IdFactory id;
		std::vector<BWAPI::Position> scout_locations;
		size_t base_index;
		
	public:

		BWAPI::Position get_next_base() {
			if(scout_locations.size() == 0) {
				return BWAPI::Positions::Invalid;
			} else {
				base_index = (base_index+1)%scout_locations.size();
				return scout_locations[base_index];
			}
		}
		void add_scouted_location(BWAPI::Position pos) {
			log("GLOBAL: Added a scouted location!");
			scout_locations.push_back(pos);
		}
		size_t get_scout_location_amount() {
			return scout_locations.size();
		}



		/* Determines wethere there is an unoccupied geyser near a Nexus */
		bool can_build_assimilator() {
			std::set<BWAPI::Unit *> geysers = BWAPI::Broodwar->getStaticGeysers();
			MyUnitGroup bases = all().mine().by_type(BWAPI::UnitTypes::Protoss_Nexus);
			
			int cnt = 0;
			for (std::set<BWAPI::Unit *>::const_iterator it = geysers.begin(); it != geysers.end(); ++it) {
				for (size_t j = 0; j != bases.size(); ++j) {
					int distance = bases.at(j)->getDistance((*it)->getPosition());
					if (distance < 400) cnt += 1;
				}
			}

			return (cnt > all().mine().by_type(BWAPI::UnitTypes::Protoss_Assimilator).size());
		}

		std::vector<BWAPI::Position> sort_by_distance(const std::vector<BWAPI::Position>& positions, const BWAPI::Position& position) {
			std::vector<BWAPI::Position> sorted;

			for (std::vector<BWAPI::Position>::const_iterator i = positions.begin(); i != positions.end(); ++i) {
				bool inserted = false;
				for (std::vector<BWAPI::Position>::iterator j = sorted.begin(); j != sorted.end(); ++j) {
					if (i->getDistance(position) < j->getDistance(position)) {
						sorted.insert(j, *i);
						inserted = true;
						break;
					}
				}
				if (!inserted)
					sorted.push_back(*i);
			}

			return sorted;
		}

		int get_supplyavailable() {
			return BWAPI::Broodwar->self()->supplyTotal() - BWAPI::Broodwar->self()->supplyUsed();
		}

		int get_supplytotal() {
			return BWAPI::Broodwar->self()->supplyTotal();
		}

		BWAPI::TilePosition find_expansion(const BWAPI::TilePosition& start_location) {
			static std::vector<BWAPI::Position> sorted;
			static size_t next = 1;
			if (sorted.size() == 0) {
				BWAPI::Position sl(start_location);
				std::set<BWTA::BaseLocation*> base_locations = BWTA::getBaseLocations();
				std::vector<BWAPI::Position> foo;
				for (std::set<BWTA::BaseLocation*>::iterator i = base_locations.begin(); i != base_locations.end(); ++i)
					foo.push_back((*i)->getPosition());
				sorted = sort_by_distance(foo, sl);
			}
			if (next == sorted.size()) return BWAPI::TilePositions::Unknown;
			return BWAPI::TilePosition(sorted[next++]);
		}

		Resource get_resources() {
			return Resource(BWAPI::Broodwar->self()->minerals(),
				BWAPI::Broodwar->self()->gas(),
				BWAPI::Broodwar->self()->supplyUsed()
			);
		}
		//TODO: update continuously, not by peeking at BWAPI :)
		void update_gamestate() {
			Gamestate state1;
			state1.frame = BWAPI::Broodwar->getFrameCount();
			state1.my_race_id = BWAPI::Broodwar->self()->getRace().getID();
			state1.their_race_id = BWAPI::Broodwar->enemy()->getRace().getID();

			state1.mMinerals = BWAPI::Broodwar->self()->minerals();
			state1.mGas = BWAPI::Broodwar->self()->gas();

			state1.mMineralsRate = resman.getMineralRate();
			state1.mGasRate = resman.getGasRate();


			MyUnitGroup all = this->all();
			
			

			state1.mHarvesters = all.mine().harvesters().size();
			state1.eHarvesters = all.hostiles().harvesters().size();

			//Our research isnt yet very interesting :P 

			//enemy research isnt yet discoverable

			MyUnitGroup mUnits = all.mine();
			mUnits.rem(mUnits.buildings());
			for(size_t i = 0 ; i < mUnits.size() ; i++) {
				Gamestate::Units unit;
				unit.count = 1;
				unit.mine = true;
				unit.type_id = mUnits.at(i)->getType().getID();
				state1.mUnits.push_back(unit);
			}
			
			MyUnitGroup eUnits = all.hostiles();
			eUnits.rem(eUnits.buildings());
			for(size_t i = 0 ; i < eUnits.size() ; i++) {
				Gamestate::Units unit;
				unit.count = 1;
				unit.mine = false;
				unit.type_id = eUnits.at(i)->getType().getID();
				state1.eUnits.push_back(unit);
			}
			
			MyUnitGroup mBuildings = all.mine().buildings();
			for(size_t i = 0 ; i < mBuildings.size() ; i++) {
				//Dont need other information that base.size atm
				if(mBuildings.at(i)->getType().getID() == BWAPI::UnitTypes::Protoss_Nexus.getID() ||
					mBuildings.at(i)->getType().getID() == BWAPI::UnitTypes::Terran_Command_Center.getID() ||
					mBuildings.at(i)->getType().getID() == BWAPI::UnitTypes::Zerg_Hatchery.getID() ||
					mBuildings.at(i)->getType().getID() == BWAPI::UnitTypes::Zerg_Lair.getID() ||
					mBuildings.at(i)->getType().getID() == BWAPI::UnitTypes::Zerg_Hive.getID()) {
						Gamestate::Bases base;
						base.mine = true;
						state1.mBases.push_back(base);
				}

				Gamestate::Units unit;
				unit.count = 1;
				unit.mine = true;
				unit.type_id = mBuildings.at(i)->getType().getID();
				state1.mBuildings.push_back(unit);
			}

			MyUnitGroup eBuildings = all.hostiles().buildings();
			for(size_t i = 0 ; i < eBuildings.size() ; i++) {
				//Dont need other information that base.size atm
				if(eBuildings.at(i)->getType().getID() == BWAPI::UnitTypes::Protoss_Nexus.getID() ||
					eBuildings.at(i)->getType().getID() == BWAPI::UnitTypes::Terran_Command_Center.getID() ||
					eBuildings.at(i)->getType().getID() == BWAPI::UnitTypes::Zerg_Hatchery.getID() ||
					eBuildings.at(i)->getType().getID() == BWAPI::UnitTypes::Zerg_Lair.getID() ||
					eBuildings.at(i)->getType().getID() == BWAPI::UnitTypes::Zerg_Hive.getID()) {
						Gamestate::Bases base;
						base.mine = false;
						state1.eBases.push_back(base);
				}
				Gamestate::Units unit;
				unit.count = 1;
				unit.mine = true;
				unit.type_id = eBuildings.at(i)->getType().getID();
				state1.eBuildings.push_back(unit);
			}
			this->state = state1;
		}
		void estimate_cost(const PlanItem& item, int& min, int& gas, int& sup) {
			min = gas = sup = 0;

			/* Not all types of planitem are listed, as some have zero cost.. */
			if (item.myType == scai::Build) {
				BWAPI::UnitType t(item.build.type_id);
				min = t.mineralPrice();
				gas = t.gasPrice();
				sup = -t.supplyProvided() / 2;
			} else if (item.myType == scai::Train) {
				BWAPI::UnitType t(item.train.type_id);
				min = t.mineralPrice() * item.train.count;
				gas = t.gasPrice() * item.train.count;
				// BWAPI counts supply as double, due to zerglings..
				sup = (t.supplyRequired() / 2)  * item.train.count;
			} else if (item.myType == scai::Upgrade) {
				/* TODO: This does not handle multi-level upgrades for now! */
				BWAPI::UpgradeType t(item.upgrade.upgrade_id);
				min = t.mineralPrice();
				gas = t.gasPrice();
			} else if (item.myType ==  scai::Expand) {
				min = 400;
			}
		}

		int estimate_time(const PlanItem& item) {
			int time = 0;
			if (item.myType == scai::Build) {
				BWAPI::UnitType t(item.build.type_id);
				time = t.buildTime();
			} else if (item.myType == scai::Train) {
				BWAPI::UnitType t(item.train.type_id);
				time = t.buildTime() * item.train.count;
			} else if (item.myType == scai::Upgrade) {
				BWAPI::UpgradeType t(item.upgrade.upgrade_id);
				time = t.upgradeTime();
			} else if (item.myType == scai::Expand) {
				time = BWAPI::UnitTypes::Protoss_Nexus.buildTime();
			}
			return time;
		}
		std::vector<BWAPI::Position> get_all_scout_locations() {
			std::vector<BWAPI::Position> ret;
			std::set<BWTA::BaseLocation*> base_locations = BWTA::getBaseLocations();

			//always check enemy location first
			ret.push_back(get_enemy_location());
			
			//Add every location that is not ours or the enemy's starting location
			for(std::set<BWTA::BaseLocation*>::const_iterator it = base_locations.begin() ; it != base_locations.end() ; it++) {
				if(get_enemy_location().getApproxDistance((*it)->getPosition()) < 500 ) {
					continue;
				}
				//If we dont have a ground path we dont want to add it
				if((*it)->getPosition().hasPath(ret[0])) {
					ret.push_back((*it)->getPosition());
				}
			}
			return ret;
		}

		BWAPI::Position get_enemy_location() {
			BOOST_FOREACH(BWAPI::TilePosition o, BWAPI::Broodwar->getStartLocations()){

				if(o.x() != BWAPI::Broodwar->self()->getStartLocation().x()){
					return BWAPI::Position(o);
				}
			}
			return BWAPI::Positions::Unknown;
		}
		
		float getMineralRate() { return resman.getMineralRate(); }
		float getGasRate() { return resman.getGasRate(); }

		ResMan& getResman() { return resman; }

		bool hasSupply(BWAPI::UnitType type) {
			return (BWAPI::Broodwar->self()->supplyUsed() + type.supplyRequired() <= BWAPI::Broodwar->self()->supplyTotal());
		}

		static std::vector<BWAPI::UnitType> direct_dependencies(const BWAPI::UnitType& type) {
			std::map<BWAPI::UnitType, int> deps = type.requiredUnits();
			std::vector<BWAPI::UnitType> ret;
			for (std::map<BWAPI::UnitType, int>::const_iterator it = deps.begin(); it != deps.end(); ++it)
				if (!(it->first.isWorker())) /* There is a circular dependency between workers and mains, break it! */
					ret.push_back(it->first);
			return ret;
		}

		/*
		 *  Returns all dependencies for a given unit or structure, ordered from most to least techy structure
		 *  IE, for Carrier:  Fleet beacon, Stargate, Cybernetics core, Gateway, Nexus
		 */
		static std::vector<BWAPI::UnitType> all_dependencies(const BWAPI::UnitType& type) {
			std::vector<BWAPI::UnitType> deps = direct_dependencies(type);

			for (std::vector<BWAPI::UnitType>::iterator dep = deps.begin(); dep != deps.end(); ++dep) {
				const std::vector<BWAPI::UnitType> current = direct_dependencies(*dep);
				for (std::vector<BWAPI::UnitType>::const_iterator it = current.begin(); it != current.end(); ++it) {
					std::vector<BWAPI::UnitType>::iterator found = std::find(deps.begin(), deps.end(), *it);
					if (found != deps.end()) {
						const BWAPI::UnitType t = *dep;
						deps.erase(found);
						dep = std::find(deps.begin(), deps.end(), t);
					}
					dep = deps.insert(dep + 1, *it) - 1;
				}
			}
			return deps;
		}

		std::vector<BWAPI::UnitType> missing_requirements(BWAPI::UnitType type) {
			std::vector<BWAPI::UnitType> ret;
			std::map<BWAPI::UnitType, int> reqs = type.requiredUnits();

			for (std::map<BWAPI::UnitType, int>::iterator it = reqs.begin(); it != reqs.end(); ++it) {
				if (all().by_type(it->first).isEmpty())
					ret.push_back(it->first);
			}
			return ret;
		}

		bool have_all(std::vector<BWAPI::UnitType> types) {
			MyUnitGroup u = all();
			for (int i = 0; i != types.size(); ++i) {
				if (u.by_type(types[i]).isEmpty())
					return false;
			}
			return true;
		}

		int getNextId(){
			return id.getNextId();
		}

		Gamestate getGamestate(){
			static int last_frame = -1;
			if (last_frame != BWAPI::Broodwar->getFrameCount())
				update_gamestate();
			return state;
		}

		BWAPI::TilePosition getBuildLocationNear(BWAPI::TilePosition pos, BWAPI::UnitType type) {
			BWAPI::TilePosition pos2 = placer.getBuildLocationNear(pos, type);
			placer.reserveTiles(pos2, type.tileWidth(), type.tileHeight());
			return pos2;
		}
		void unReserve (BWAPI::TilePosition pos, int width, int height) {
			placer.freeTiles(pos, width, height);
		}


		BWAPI::TilePosition startLocation() {
			return BWAPI::Broodwar->self()->getStartLocation();
		}

		/**
		 *  Return the amount of minerals which are currently available to use.
		 */
		int getAvailableMinerals() {
			return BWAPI::Broodwar->self()->minerals() - committedMinerals;
		}


		/**
		 *  Return the amount of gas available to you now.
		 */
		int getAvailableGas() {
			return BWAPI::Broodwar->self()->gas() - committedGas;
		}


		/**
		 *  Use this to mark some minerals as spent
		 *  Remember to un-commit them later!
		 */
		void commitMinerals(int minerals) {
			committedMinerals += minerals;
		}


		/**
		 *  Use this to mark minerals as used.
		 */
		void unCommitMinerals(int minerals) {
			committedMinerals -= minerals;
		}


		/**
		 *  Use this to mark some gas as spent
		 *  Remember to un-commit them later!
		 */
		void commitGas(int gas) {
			committedGas += gas;
		}

		/**
		 *  Use this to mark gas as used.
		 */
		void unCommitGas(int gas) {
			committedGas -= gas;
		}


		/**
		 *  Get the current frame of the game
		 */
		int getCurrentFrame() {
			return BWAPI::Broodwar->getFrameCount();
		}


		/**
		 *  Get a group of _all_ units currently in the game.
		 */
		MyUnitGroup all() {
			std::set<BWAPI::Unit*> u = BWAPI::Broodwar->getAllUnits();
			MyUnitGroup a(std::vector<BWAPI::Unit*>(u.begin(), u.end()));
			a.remove_dead();
			return a;
		}

		static Global* i() {
			static Global instance;
			return &instance;
		}
	};
}