#pragma once

#include "Communication.h"
#include "MyUnitGroup.h"

namespace scai {
	/*
	 TODO: Handle special cases:
	 1. Gas nodes get harrassed, must be filled again
	 */
	struct HarvestManager {

		struct ResourceLocation {
			BWAPI::Position location;
			MyUnitGroup resources;
			MyUnitGroup harvesters;
			int behaviour_id;
			bool isGas;
		};

		ILayer * executive;
		Communication& comm;
		std::vector<ResourceLocation> locations;

		HarvestManager(Communication& comm): comm(comm), executive(NULL) {}

		void add_resource_location(BWAPI::Position location) {
			ResourceLocation next;

			next.location = location;
			int range = 320;
			
			next.resources = scai::Global::i()->all().minerals().withinRange(location, range);
			log("addresourcestuff range is %d, found %d patches", range, next.resources.size());
			while(next.resources.size() == 0) {
				range+=320;
				next.resources = scai::Global::i()->all().minerals().withinRange(location, range);
				log("Trying again, range is now %d, found %d patches", range, next.resources.size());
			}
			next.harvesters = split_harvesters();
			next.behaviour_id = scai::Global::i()->getNextId();
			next.isGas = false;
			
			spawn_behaviour(next.harvesters, next.resources, next.behaviour_id);

			locations.push_back(next);
		}

		void add_gas_location(BWAPI::Position location, MyUnitGroup assimilator) {
			ResourceLocation next;

			next.location = location;
			next.resources = assimilator;
			next.harvesters = grab_gassers(3);
			next.behaviour_id = scai::Global::i()->getNextId();
			next.isGas = true;
			
			spawn_behaviour(next.harvesters, next.resources, next.behaviour_id);

			locations.push_back(next);
		}

		MyUnitGroup grab_gassers(size_t amount) {
			MyUnitGroup temp;
			for( size_t i = 0 ; i < locations.size() ; i++) {
				if(!locations[i].isGas && locations[i].harvesters.size() > 2) {
					BWAPI::Unit* u = locations[i].harvesters.at(0);
					BWAPI::Unit* u2 = locations[i].harvesters.at(1);
					BWAPI::Unit* u3 = locations[i].harvesters.at(2);
					rem(u);
					rem(u2);
					rem(u3);
					temp.add(u);
					temp.add(u2);
					temp.add(u3);
				}

			}
			return temp;
		}

		MyUnitGroup split_harvesters() {
			//we need to collect 5 initial harvesters from the other locations, given that they have additional
			MyUnitGroup temp;
			for(size_t i = 0 ; i < 5 ; i++) {
				size_t index = -1;
				size_t max = 5;
				for(size_t j = 0 ; j < locations.size() ; j++) {
					if(locations[j].harvesters.size() > max) {
						index = j;
						max = locations[j].harvesters.size();
					}
				}
				if (index != -1) {
					log("Adding a harvester from a group at index %d", index);
					BWAPI::Unit* u = locations[index].harvesters.at(0);
					log("Finished adding");
					rem(u);
					temp.add(u);
				}
			}
			return temp;
		}

		void spawn_behaviour(const MyUnitGroup& harvesters, const MyUnitGroup& resources, int id) {
			Message harvest;
			harvest.id = NewBehaviour;
			harvest.putString("behaviour", "harvest");
			harvest.put("harvesters", harvesters);
			harvest.put("minerals", resources);
			harvest.put("behaviourId", id);
			comm.sendDn(executive, harvest);
		}


		void add(BWAPI::Unit* unit) {
			//Add the worker to the nearest location, but if it is saturated, go to the one with the least workers
			
			int index = getClosestUnsaturated(unit->getPosition());
			locations[index].harvesters.add(unit);

			Message add;
			add.id = scai::AddUnits;
			add.put("behaviour_id", locations[index].behaviour_id);
			add.put("units", MyUnitGroup(unit));
			comm.sendDn(executive, add);
		}

		size_t getClosestUnsaturated(BWAPI::Position pos) {
			int index = -1;

			int distance = 100000000;
			for(size_t i = 0 ; i < locations.size() ; i++) {
				if (index == -1 && !locations[i].isGas) {
					index = i;
					distance = locations[i].location.getApproxDistance(pos);
				}
				else if( distance > locations[i].location.getApproxDistance(pos) && !locations[i].isGas && locations[index].harvesters.size() < locations[index].resources.size()*3) {
					index = i;
					distance = locations[i].location.getApproxDistance(pos);
				}

			}
			return index;
		}

		MyUnitGroup getWorker() {
			for(size_t i = 0 ; i < locations.size() ; i++) {
				if (!locations[i].isGas && locations[i].harvesters.size() > 0) {
					BWAPI::Unit* temp = locations[i].harvesters.at(0);
					rem(temp);
					return MyUnitGroup(temp);
				}
			}
			return MyUnitGroup();
		}

		void rem(BWAPI::Unit* unit) {
			for(size_t i = 0 ; i < locations.size() ; i++) {
				if(locations[i].harvesters.contains(unit)){
					locations[i].harvesters.rem(unit);
					break;
				}
			}

			Message rem;
			rem.id = scai::ReleaseUnit;
			rem.put("unit", MyUnitGroup(unit));
			comm.sendDn(executive, rem);
		}

		void onStart() {
			BWAPI::Position pos(BWAPI::Broodwar->self()->getStartLocation());
			
			add_resource_location(pos);
		}

		void onFrame() {
			for (size_t i = 0; i != locations.size(); ++i)
				locations[i].harvesters.remove_dead();

		}
	};
}