#pragma once

#include <string>
#include <map>
#include <boost/shared_ptr.hpp>
#include <boost/any.hpp>
#include "MessageTypes.h"
#include "log.h"




namespace scai {

	typedef std::map<std::string, boost::any> Payload;
	class Message {
	public:
		Message(): payload(new Payload()) {}

		MessageId id;

		boost::shared_ptr<Payload> payload;

		bool operator == (const Message& o) {
			return (id == o.id) && (payload == o.payload);
		}

		boost::any get(std::string key) {
			try {
				return (*payload)[key];
			} catch (std::exception e) {
				log ("ERROR when getting key %s", key.c_str());
				throw e;
			}
		}
		void put(std::string key, boost::any val) {
			(*payload)[key] = val;
		}

		template <typename T>
		T getCast(std::string key) {
			try {
				return boost::any_cast<T>(get(key));
			} catch (boost::bad_any_cast e) {
				log ("Bad cast with key %s", key.c_str());
				throw e;
			}
		}

		std::string getString(std::string key) {
			try {
				return boost::any_cast<std::string>(get(key));
			} catch (boost::bad_any_cast e) {
				log ("Bad string cast with key %s", key.c_str());
				throw e;
			}
		}

		void putString(std::string key, std::string value) {
			put(key, value);
		}
	};
}