namespace scai{
	enum MessageId {
		/*Expected key "plan" and payload Plan
		Will be a message from planning layer to executive with a new plan*/	
		NewPlan,

		/**
		 *  Executive -> Reactive
		 *  Signals the reactive layer to instantiate a new behaviour.
		 *  Expects "behaviour" -> std::string representing the behaviour to spawn.
		 *  Expects additional keys according to behaviour.
		 */
		NewBehaviour,

		/**
		 *  Executive -> Reactive
		 *  The reactive layer should release a unit from its current behaviour
		 *  Expects "unit" -> BWAPI::Unit*
		 */
		ReleaseUnit,

		/**
		 *  Reactive -> Executive
		 *  A behaviour in the reactive layer has released a unit, the executive should deallocate it.
		 *
		 *  Expects "unit" -> MyUnitGroup, the units to release
		 */
		RemoveUnit,

		/**
		 *  Executive -> Reactive
		 *  The reactive layer should add a group of units to a given behaviour.
		 *
		 *  Expects:
		 *  "units" -> MyUnitGroup
		 *  "behaviour_id" -> int
		 */
		AddUnits,

		PlanStatus,

		
		

		/*Expected key "planFinished", and payload planId
		Will be a message from executive to planning when a plan has been finished 
		so it knows when to replan*/
		PlanFinished,

		TestItem,


		/**
		 * Reactive (individual behaviour) -> Executive
		 * The behaviour in the reactive layer should send a confirmation message that it has succeeded
		 * Expects key "behaviourStatus", value can be "Success" or "Failure"
		 * Expects empty payload, but a set behaviourId in the message
		 */
		BehaviourStatus,

		/**
		 *Reactive -> Executive
		 * meh!
		 */
		BuildingCompleted,

		/**
		 * Reactive -> Executive
		 * Sent when scout finds a new enemy base location
		 */
		ScoutLocation
	};
}