#pragma once
#include "PlanItem.h"
#include "Gamestate.h"
#include <boost/foreach.hpp>
#include <BWAPI.h>
#include "CBRModule.h"
#include "Case.h"
#include "Global.h"
#include "TestRunner.h"

namespace scai {
	class PlanGenerator {
	private: 
		//TODO: temporary for progressing in plans!
		int planCount;
		CBRModule cbr;
		scai::TestRunner* test_runner;


		public:
			PlanGenerator(const std::string& dbindex, TestRunner* runner): planCount(0), cbr(dbindex), test_runner(runner) {}
		/**
		* Currently hardcoded
		* Needs to use gamestate to find gametime and locations in the future in the future
		*/
			void addTrain(Plan& plan, BWAPI::UnitType unitType, int count){
				PlanItem temp;
				temp.myType = scai::Train;
				temp.train.type_id = unitType.getID();
				temp.train.count = count;
				plan.push_back(temp);
			}
			void addBuilding(Plan& plan, BWAPI::UnitType unitType){
				PlanItem temp;
				temp.myType = scai::Build;
				temp.build.type_id = unitType.getID();
				plan.push_back(temp);
			}

			void addAttack(Plan& plan, BWAPI::Position pos){
				PlanItem temp;
				temp.myType = scai::Attack;
				temp.attack.location = scai::Main;
				plan.push_back(temp);
			}

			void addDefend(Plan& plan, BWAPI::Position pos){
				PlanItem temp;
				temp.myType = scai::Defend;
				temp.defend.location = scai::Main;
				plan.push_back(temp);
			}

			void addScout(Plan& plan){
				PlanItem temp;
				temp.myType = scai::Scout;
				temp.scout.location = scai::Main;
				plan.push_back(temp);
			}

			void addExpand(Plan& plan, BWAPI::Position pos){
				PlanItem temp;
				temp.myType = scai::Expand;
				temp.expand.location = scai::Expansion;
				plan.push_back(temp);
			}
			void addUpgrade(Plan& plan, BWAPI::UpgradeType upgrade){
				PlanItem temp;
				temp.myType = scai::Upgrade;
				temp.upgrade.upgrade_id = upgrade.getID();
				plan.push_back(temp);
			}


			void print_match(const Case& case1, const Gamestate& current ) {
				log("Logging for case with name %s", case1.match_id.c_str());
				log("Printing our gamestate:");
				print_gamestate(current);
				log("Printing case gamestate");
				print_gamestate(case1.state);
				log("Printing plan:");

				for(size_t i = 0 ; i < case1.plan.size() ; i++) {
					PlanItem item = case1.plan[i];
					switch(item.myType) {
						case scai::Build: 
							log("Building: with id %d", item.build.type_id);
							break;
						case scai::Train: 
							log("Train: with id %d", item.train.type_id);
							break;
						case scai::Upgrade: 
							log("Upgrade: with id %d", item.upgrade.upgrade_id);
							break;
						case scai::Attack: 
							log("Attack!");
							break;
						case scai::Defend:
							log("DEFEND!");
							break;
						case scai::Expand: 
							log("EXPAND!");
							break;
						case scai::Scout: 
							log("SCOUT!");
							break;
						default:
							log("DUDE ERROR!!!");
							break;
					}
				}
			}
			
			virtual Plan generatePlan(){
				float match = -1.0;
				test_runner->start_case();
				scai::Global* g = scai::Global::i();
				Gamestate current = g->getGamestate();
				Case case1 = cbr.get_case(current, match);
				test_runner->end_case(case1, current, match);

				print_match(case1, current);

				//TODO: This is a hack to ensure that we scout on the first case
				if(BWAPI::Broodwar->getFrameCount() < 200 && case1.plan[case1.plan.size()-1].myType != scai::Scout) {
					PlanItem it;
					it.myType = scai::Scout;
					it.scout.location = scai::Main;
					if(case1.plan.size() > 4) {
						case1.plan.insert(case1.plan.begin()+4, it);
					} else {
						case1.plan.push_back(it);
					}
				}
				/*//This is a current hack to make archons and stuff :)
				Plan meh;
				addTrain(meh, BWAPI::UnitTypes::Protoss_Probe, 4);
				addBuilding(meh, BWAPI::UnitTypes::Protoss_Pylon);
				addTrain(meh, BWAPI::UnitTypes::Protoss_Probe, 5);
				addBuilding(meh, BWAPI::UnitTypes::Protoss_Assimilator);
				addBuilding(meh, BWAPI::UnitTypes::Protoss_Pylon);
				addBuilding(meh, BWAPI::UnitTypes::Protoss_Gateway);
				addBuilding(meh, BWAPI::UnitTypes::Protoss_Cybernetics_Core);
				addBuilding(meh, BWAPI::UnitTypes::Protoss_Citadel_of_Adun);
				addBuilding(meh, BWAPI::UnitTypes::Protoss_Templar_Archives);
				addTrain(meh, BWAPI::UnitTypes::Protoss_High_Templar, 2);
				addScout(meh);
				case1.plan = meh;*/

				return case1.plan;
			}

			void print_gamestate(Gamestate state) {

				log("For state!");
				log("frame: %d, mins: %d, gas: %d, minrate: %f, gasrate: %f, harvester: %d, eHarv: %d", state.frame, state.mMinerals, state.mGas, state.mMineralsRate, state.mGasRate, state.mHarvesters, state.eHarvesters);
				log("Research:");
				for(std::map<int, int>::const_iterator it = state.mResearch.begin() ; it != state.mResearch.end() ; ++it) {
					log("For research %d, lvl %d", it->first,  it->second);
				}
				log("eResearch:");
				for(std::map<int, int>::const_iterator it = state.eResearch.begin() ; it != state.eResearch.end() ; ++it) {
					log("For research %d, lvl %d", it->first,  it->second);
				}
				log("My units");
				for(size_t i = 0 ; i < state.mUnits.size() ; i++) {
					log("Unit at %d, count %d, type %d", i, state.mUnits[i].count, state.mUnits[i].type_id);
				}

				log("e units");
				for(size_t i = 0 ; i < state.eUnits.size() ; i++) {
					log("Unit at %d, count %d, type %d", i, state.eUnits[i].count, state.eUnits[i].type_id);
				}

				log("My buildings");
				for(size_t i = 0 ; i < state.mBuildings.size() ; i++) {
					log("Unit at %d, count %d, type %d", i, state.mBuildings[i].count, state.mBuildings[i].type_id);
				}

				log("e buildings");
				for(size_t i = 0 ; i < state.eBuildings.size() ; i++) {
					log("Unit at %d, count %d, type %d", i, state.eBuildings[i].count, state.eBuildings[i].type_id);
				}
				log("Bases size: %d %d", state.mBases.size(), state.eBases.size());
			}

			//dummy implementation
			virtual bool validatePlan(Gamestate state, Plan plan){
				return true;
			}
	

	

	};
}