#pragma once

#include "ILayer.h"
#include "log.h"
#include "BuildingPlacer.h"
#include "IBehaviour.h"
#include "BaseBehaviours.h"
#include "RootBehaviour.h"

#include <BWAPI.h>
#include <BWTA.h>

/*
 Note to self:  Crashes on 8 probes.  WHY??
*/

namespace scai {

	
	
	class Reactive : public ILayer {
		Communication & comm;
		RootBehaviour root;
		std::vector<Message> msgBuffer;

	public:
		
		virtual void onFrame() {
			while (comm.hasNext(this)) {
				Message msg = comm.getNext(this);

				if (msg.id == ReleaseUnit) {
					MyUnitGroup group = msg.getCast<MyUnitGroup>("unit");
					root.releaseUnits(group);
				} else if (msg.id == NewBehaviour) {
					if (msg.getString("behaviour") == "train") {
						spawn_train(msg.getCast<BWAPI::UnitType>("unit"), msg.getCast<MyUnitGroup>("producer"), msg.getCast<int>("behaviourId"));
					} else if (msg.getString("behaviour") == "harvest") {
						spawn_harvest(msg.getCast<MyUnitGroup>("harvesters"),
									  msg.getCast<MyUnitGroup>("minerals"),
									  msg.getCast<int>("behaviourId"));
					} else if (msg.getString("behaviour") == "build") {
						MyUnitGroup builder = msg.getCast<MyUnitGroup>("worker");
						BWAPI::Position pos = msg.getCast<BWAPI::Position>("location");
						spawn_build(msg.getCast<BWAPI::UnitType>("type"), pos, builder, msg.getCast<int>("behaviourId"));
					} else if (msg.getString("behaviour") == "attackBase") {
						BWAPI::Broodwar->printf("SPAWN attack behviour");
						MyUnitGroup attackers = msg.getCast<MyUnitGroup>("unit");
						BWAPI::Position position = msg.getCast<BWAPI::Position>("position");

						TRACE( spawn_attack(attackers, position, msg.getCast<int>("behaviourId")), "Attack");
					} else if (msg.getString("behaviour") == "scoutBaseLocation") {
						MyUnitGroup scouter = msg.getCast<MyUnitGroup>("unit");
						std::vector<BWAPI::Position> pos = msg.getCast<std::vector<BWAPI::Position>>("position");

						spawn_scout(scouter, pos, msg.getCast<int>("behaviourId"));
					} else if (msg.getString("behaviour") == "expand") {
						spawn_expand(msg.getCast<MyUnitGroup>("worker"), msg.getCast<BWAPI::TilePosition>("location"), msg.getCast<int>("behaviourId"));
					} else if(msg.getString("behaviour") == "moveto") {
						spawn_move(msg.getCast<MyUnitGroup>("unit"), msg.getCast<BWAPI::Position>("position"), msg.getCast<int>("behaviourId"));
					} else if(msg.getString("behaviour") == "archon") {
						spawn_archon(msg.getCast<MyUnitGroup>("unit"), msg.getCast<int>("behaviourId"));
					}
					
				} else if (msg.id == AddUnits) {
					root.addUnit(msg.getCast<MyUnitGroup>("units"), msg.getCast<int>("behaviour_id"));
				} else {
					log("WARN: Reactive: Unknown message id");
					BWAPI::Broodwar->printf("WARN: Reactive: Unknown message id");
				}
			}
			
			root.doFrame();
			while(root.hasMoreMessages()) {
				
				Message r = root.retrieveMessage();
				comm.sendUp(this, r);
			}
			root.print_behaviours(440, 25);
		}

		Reactive(Communication & comm): comm(comm) {}

		void spawn_expand(const MyUnitGroup& worker, const BWAPI::TilePosition& location, int behaviour_id) {
			IBehaviour* expand = new scai::ExpandBehaviour(worker, location);
			expand->behaviourId = behaviour_id;
			root.addChild(expand);
		}

		void spawn_attack(const MyUnitGroup& attackers, BWAPI::Position pos, int behaviourId) {
			//Find position slightly away from ramp
			BWAPI::Position pos2;

			BWTA::Region* reg;
			BWTA::Chokepoint* choke;
			
			reg = BWTA::getRegion(BWAPI::TilePosition(pos));
			choke = *reg->getChokepoints().begin();


			//let the units move to halfway the distance from the attack position and the adjoining ramp from the ramp
			pos2.x() = choke->getCenter().x() + (choke->getCenter().x() - pos.x())/2;
			pos2.y() = choke->getCenter().y() + (choke->getCenter().y() - pos.y())/2;

			IBehaviour* move = new scai::AttackMoveToBehaviour(attackers, pos2);
			
			IBehaviour* attack = new scai::AttackBehaviour(attackers, pos);
			attack->behaviourId = behaviourId;

			IBehaviour* wrapper = new scai::ReaverWrapper(attackers, attack);
			std::vector<IBehaviour*> children;
			children.push_back(move);
			children.push_back(wrapper);
				
			IBehaviour* sequence = new scai::Sequence(attackers, children);
			TRACE(root.addChild(sequence), "AddChild");
		}

		void spawn_build(BWAPI::UnitType type, const BWAPI::Position& location, const MyUnitGroup& builder, int behaviourId) {
			BWAPI::Position pos;
			if (location == BWAPI::Positions::Unknown)
				pos = BWAPI::Position(scai::Global::i()->getBuildLocationNear(scai::Global::i()->startLocation(), type));
			else
				pos = location;
			
			std::vector<IBehaviour*> c;
			c.push_back(new scai::MoveToBehaviour(builder, BWAPI::Position(pos)));
			c.push_back(new scai::BuildBehaviour(builder, BWAPI::Position(pos), type));
			IBehaviour* seq = new scai::Sequence(builder, c);
			seq->behaviourId = behaviourId;

			root.addChild(seq);
		}

		void spawn_train(BWAPI::UnitType type, const MyUnitGroup& producer, int behaviourId) {
			IBehaviour* train = new scai::TrainBehaviour(producer, type);
			train->behaviourId = behaviourId;
			root.addChild(train);
		}

		void spawn_harvest(const MyUnitGroup& harvesters, const MyUnitGroup& minerals, int behaviour_id) {
			IBehaviour* harvest = new scai::HarvestBehaviour(harvesters, minerals);
			harvest->behaviourId = behaviour_id;
			root.addChild(harvest);
		}

		void spawn_scout(const MyUnitGroup& scouter, std::vector<BWAPI::Position> pos, int behaviour_id) {
			IBehaviour* scout = new scai::ScoutBehaviour(scouter, pos);
			scout->behaviourId = behaviour_id;
			root.addChild(scout);
		}

		void spawn_move(const MyUnitGroup& units, BWAPI::Position pos, int behaviour_id) {
			IBehaviour* move = new scai::AttackMoveToBehaviour(units, pos);
			move->behaviourId = behaviour_id;
			root.addChild(move);
		}

		void spawn_archon(const MyUnitGroup& units, int behaviour_id) {
			IBehaviour* archon = new scai::ArchonBehaviour(units);
			archon->behaviourId = behaviour_id;
			root.addChild(archon);
		}
	};
}