#pragma once

#include <BWAPI.h>
#include "IBehaviour.h"

namespace scai {
	struct RootBehaviour : public IBehaviour {
		std::vector<IBehaviour *> remove;
		RootBehaviour(): IBehaviour(MyUnitGroup()) {}

		void addChild(IBehaviour* child) {
			child->onStart();
			children.push_back(child);
		}

		void remChild(IBehaviour* child) {
			std::vector<IBehaviour*>::iterator it = std::find(children.begin(), children.end(), child);

			if (it == children.end()) {
				log("WARNING: root behaviour could not find the given child");
			} else {
				for (int i = 0; i != child->messages.size(); ++i) {
					messages.push_back(child->messages[i]);
				}
				child->messages.clear();
				children.erase(it);
			}
		}

		virtual Response onFrame() {
			for (int i = 0; i != children.size(); ++i) {
				Response r = children[i]->doFrame();
				if (r.status == FAILURE) onChildFailure(children[i]);
				else if (r.status == SUCCESS) onChildSuccess(children[i]);
			}

			int s = remove.size();
			for (int i = s - 1; i >= 0; --i) {
				remChild(remove[i]);
			}
			remove.clear();
			return Response(EXECUTING);
		}

		//TODO: Currently using planstatus for behaviours, change when it works again :)
		virtual Response onChildFailure(IBehaviour* child) {
			//Add a message for reactive to send to executive
			if(child->behaviourId != -1){
				Message failure;
				failure.id = scai::BehaviourStatus;
				failure.put("behaviourId", child->behaviourId);
				failure.putString("behaviourStatus", "Failure");
				messages.push_back(failure);
			}
			remove.push_back(child);

			return Response(EXECUTING);
		}

		//TODO: Currently using planstatus for behaviours, change when it works again :)
		virtual Response onChildSuccess(IBehaviour* child) {
			//Add a message for reactive to send to executive
			log("child was successful!");
			if(child->behaviourId != -1){
				log("Creating message to be sent for behaviourId %d", child->behaviourId);
				Message success;
				success.id = scai::BehaviourStatus;
				success.put("behaviourId", child->behaviourId);
				success.putString("behaviourStatus", "Success");
				messages.push_back(success);
			}
			remove.push_back(child);
			return Response(EXECUTING);
		}

		//TODO: Fix so it doesnt retrieve if there are no messages, currently safely handled in reactive though
		Message retrieveMessage() {
			Message r = messages.front();
			messages.erase(std::find(messages.begin(), messages.end(), r));
			return r;
		}

		bool hasMoreMessages() {
			if(messages.size()){
				return true;
			}
			else {
				return false;
			}
		}

		void addUnit(const MyUnitGroup& units, int behaviour_id) {
			for (int i = 0; i != children.size(); ++i) {
				if (children[i]->behaviourId == behaviour_id) {
					log ("adding unit to behaviour %s", children[i]->name());
					children[i]->addUnits(units);
					break;
				}
			}
		}
	};
}