#pragma once


#include <BWAPI.h>
#include "Gamestate.h"
#include "Case.h"
#include "CaseMatcher.h"
#include "log.h"
#include "Planitem.h"
#include "Database.h"
#include "CompositeCase.h"
#include "CBRModule.h"
#include "TestRunner.h"
#include "util.h"

#include <time.h>
#include <Windows.h>

namespace scai {
	class TestClass {
	public:
		void runTests() {
			//caseMatchTest();
			//compositeTest();
			test_dependencies();
			test_stringstuff();
			test_stringstuff2();
			test_ini();
			test_path();
			test_filesplit();
		}
	private:
		void test_filesplit() {
			std::string fname("foo.bar.baz");
			util::file_ext f = util::file_split(fname);
			scai::log("File: '%s', ext: '%s'", f.first.c_str(), f.second.c_str());

			fname = "foobar";
			f = util::file_split(fname);
			scai::log("File: '%s', ext: '%s'", f.first.c_str(), f.second.c_str());
		}

		void test_path() {
			std::string path("c:\\foo\\bar\\baz.bat");
			util::path_pair s = util::path_split(path);
			scai::log("Path: %s, File: %s", s.first.c_str(), s.second.c_str());

			path = "c:\\foo\\bar";
			s = util::path_split(path);
			scai::log("Path: %s, File: %s", s.first.c_str(), s.second.c_str());

			path = "c:\\foo\\bar\\";
			s = util::path_split(path);
			scai::log("Path: %s, File: %s", s.first.c_str(), s.second.c_str());

			path = "foo.bar";
			s = util::path_split(path);
			scai::log("Path: %s, File: %s", s.first.c_str(), s.second.c_str());
		}

		void test_ini() {
			scai::log("reading ini");
			BwapiIni ini;
			try {
				ini.load("bwapi-data\\bwapi.ini");
			} catch (std::exception e) {
				scai::log("%s", e.what());
				throw e;
			}

			scai::log("writing ini");
			ini.save("bwapi-data\\test.ini");
			scai::log("... done");
		}

		void test_stringstuff2() {
			scai::log("testing strings 2");

			std::string foo("key = value");
			std::string tail = util::split(foo, '=');
			scai::log("'%s'  '%s'", foo.c_str(), tail.c_str());

			foo = "key value";
			tail = util::split(foo, '=');
			scai::log("'%s'  '%s'", foo.c_str(), tail.c_str());

			foo = "";
			tail = util::split(foo, '=');
			scai::log("'%s'  '%s'", foo.c_str(), tail.c_str());

			scai::log("... done");
		}

		void test_stringstuff() {
			scai::log("testing strings");

			std::string foo = "    hello world  ";
			std::string bar = util::strip(foo);
			scai::log("'%s'", bar.c_str());
			
			std::string baz = "    ";
			bar = util::strip(baz);
			scai::log("'%s'", bar.c_str());

			std::string hoo = "";
			bar = util::strip(hoo);
			scai::log("'%s'", bar.c_str());

			scai::log("... done");
		}

		void test_dependencies() {
			scai::log("Testing dependencies");
			std::vector<BWAPI::UnitType> d1 = scai::Global::all_dependencies(BWAPI::UnitTypes::Protoss_High_Templar);
			scai::log("Results:");
			scai::log("%s", d1[0].getName().c_str());
			for (size_t i = 1; i < d1.size(); ++i)
				scai::log("%s", d1[i].getName().c_str());
			scai::log("Done");
		}

		void caseMatchTest() {
			CaseMatcher match;
			Gamestate state;
			state.my_race_id = BWAPI::Races::Protoss.getID();
			state.their_race_id = BWAPI::Races::Zerg.getID();
			state.mResearch[BWAPI::UpgradeTypes::Protoss_Ground_Weapons.getID()] = 1;
			state.mResearch[BWAPI::UpgradeTypes::Protoss_Air_Weapons.getID()] = 1;

			state.eResearch[BWAPI::UpgradeTypes::Zerg_Melee_Attacks.getID()] = 1;

			scai::Gamestate::Bases base;
			base.mine = true;
			base.type = scai::Main;
			state.mBases.push_back(base);

			base.type = scai::Expansion;
			state.mBases.push_back(base);

			base.mine = false;
			state.eBases.push_back(base);

			scai::Gamestate::Units unit;
			unit.mine = true;
			unit.count = 5;
			unit.type_id = BWAPI::UnitTypes::Protoss_Probe.getID();
			state.mUnits.push_back(unit);

			unit.type_id = BWAPI::UnitTypes::Protoss_Zealot.getID();
			state.mUnits.push_back(unit);

			unit.mine = false;
			unit.type_id = BWAPI::UnitTypes::Zerg_Drone.getID();
			state.eUnits.push_back(unit);

			unit.type_id = BWAPI::UnitTypes::Zerg_Zergling.getID();
			state.eUnits.push_back(unit);

			scai::Gamestate::Units building;
			building.mine = true;
			building.count = 1;
			building.type_id = BWAPI::UnitTypes::Protoss_Nexus.getID();
			state.mBuildings.push_back(building);

			building.type_id = BWAPI::UnitTypes::Protoss_Pylon.getID();
			state.mBuildings.push_back(building);

			building.mine = false;
			building.type_id = BWAPI::UnitTypes::Zerg_Hatchery.getID();
			state.eBuildings.push_back(building);



			state.map = "Astral Balance";
			state.frame = 0;


			state.mMinerals = 100;
			state.mMineralsRate = 0.36f;
			state.mGas = 10;
			state.mGasRate = 0.12f;

			state.mHarvesters = 5;
			state.eHarvesters = 5;

			PlanItem testItem;
			testItem.myType = scai::Build;
			testItem.build.type_id = BWAPI::UnitTypes::Protoss_Pylon.getID();

			PlanItem testItem2;
			testItem2.myType = scai::Train;
			testItem2.train.type_id = BWAPI::UnitTypes::Protoss_Probe.getID();
			testItem2.train.count = 4;

			PlanItem testItem3;
			testItem3.myType = scai::Upgrade;
			testItem3.upgrade.upgrade_id = BWAPI::UpgradeTypes::Protoss_Air_Weapons.getID();

			PlanItem testItem4;
			testItem4.myType = scai::Attack;
			testItem4.attack.location = scai::Main;

			PlanItem testItem5;
			testItem5.myType = scai::Defend;
			testItem5.defend.location = scai::Army;


			PlanItem testItem6;
			testItem6.myType = scai::Expand;
			testItem6.expand.location = scai::Expansion;

			PlanItem testItem7;
			testItem7.myType = scai::Scout;
			testItem6.scout.location = scai::Default;


			Plan actions;
			actions.push_back(testItem);
			actions.push_back(testItem2);
			actions.push_back(testItem3);
			actions.push_back(testItem4);
			actions.push_back(testItem5);
			actions.push_back(testItem6);
			actions.push_back(testItem7);

			Plan opponentmodel;
			testItem.build.type_id = BWAPI::UnitTypes::Zerg_Spawning_Pool.getID();
			opponentmodel.push_back(testItem);
			testItem2.train.type_id = BWAPI::UnitTypes::Zerg_Drone.getID();
			opponentmodel.push_back(testItem2);

			Case testcase = Case(state, actions, opponentmodel, 100, 110);
			

			log("Matching case: %f",match.plannerCompare(testcase, state, 0.0f));
			
			state.mMinerals = 110;
			state.mMineralsRate = 0.40f;
			state.mGas = 10;
			state.mGasRate = 0.12f;

			state.mHarvesters = 6;
			state.eHarvesters = 7;

			state.mResearch[BWAPI::UpgradeTypes::Protoss_Ground_Weapons.getID()] = 2;

			testcase.state.eResearch[BWAPI::UpgradeTypes::Protoss_Ground_Weapons.getID()] = 2;

			base.mine = true;
			base.type = scai::Expansion;
			state.mBases.push_back(base);

			building.mine = true;
			building.count = 1;
			building.type_id = BWAPI::UnitTypes::Protoss_Gateway.getID();
			state.mBuildings.push_back(building);

			
			testcase.state.mBuildings.push_back(building);

			testcase.state.mBuildings.push_back(building);

			

			unit.mine = true;
			unit.count = 1;
			unit.type_id = BWAPI::UnitTypes::Protoss_Probe.getID();
			state.mUnits.push_back(unit);

			unit.mine = false;
			unit.count = 2;
			unit.type_id = BWAPI::UnitTypes::Protoss_Probe.getID();
			testcase.state.mUnits.push_back(unit);



			log("Matching mismatched case: %f",match.plannerCompare(testcase, state, 0.0f));

			//Testing database here!
			/*Database db;

			CaseLibrary lib;
			for(size_t i = 0 ; i < 6000 ; i++) {
				lib.push_back(testcase);
				lib.push_back(testcase);
			}
			
			log("Pre write");
			db.write_Database(lib);


			CaseLibrary testlib;
			log("Pre read");
			testlib = db.read_Database();
			log("post read, library consists of %d elements", testlib.size());

			clock_t t1,t2;
			t1=clock();
			//code goes here
			db.getPlan(state);
			t2=clock();
			float diff ((float)t2-(float)t1);

			log("Time it took: %f", diff);


			if(testlib.size() > 0) {
				log("Matching read case from database with testcase, equality %f %f", match.plannerCompare(testlib.at(0), testcase.state, 0.0f), match.plannerCompare(testlib.at(1), testcase.state, 0.0f));
			}
			else {
				log("Read library doesnt contain the item it should!");
			}*/



		}


	};
}
