#pragma once

#include <BWAPI.h>
#include <vector>
#include <map>
#include "HarvestManager.h"

namespace scai {
	struct UnitManager {
		static bool isMyUnit(BWAPI::Unit* unit) {
			return (
				(unit->getPlayer() == BWAPI::Broodwar->self()) &&
				(!unit->getType().isBuilding()) &&
				(!unit->isHallucination())
			);
		}

		void print_allocations(int x, int y) {
			BWAPI::Broodwar->drawTextScreen(x, y, "X: %d", units.size());
			int i = 1;
			for (std::map<int, MyUnitGroup>::iterator it = allocations.begin(); it != allocations.end(); ++it) {
				BWAPI::Broodwar->drawTextScreen(x, y + 8 * i, "%d: %d", it->first, it->second.size());
				++i;
			}
		}

		HarvestManager& hm;
		MyUnitGroup under_construction;
		MyUnitGroup units;
		MyUnitGroup newly_created;
		std::map<int, MyUnitGroup> allocations;

		UnitManager(HarvestManager& hm): hm(hm) {}

		void onUnitCreate(BWAPI::Unit* unit) {
			if (isMyUnit(unit)) {
				if (!unit->isCompleted()) {
					under_construction.add(unit);
				} else {
					units.add(unit);
				}
			}
		}

		void allocate(const MyUnitGroup& alloc, int id) {
			allocations[id].add(alloc);
			units.rem(alloc);
		}

		void deallocate_all(int id) {
			units.add(allocations[id]);
			allocations.erase(id);
		}

		void deallocate_any(const MyUnitGroup& units) {
			for (std::map<int, MyUnitGroup>::iterator it = allocations.begin(); it != allocations.end(); ++it) {
				it->second.rem(units);
				if (it->second.isEmpty())
					it = allocations.erase(it);
				if (it == allocations.end()) break;
			}
		}

		void deallocate(const MyUnitGroup& dealloc, int id) {
			units.add(dealloc);
			allocations[id].rem(dealloc);
		}

		void add_units(const MyUnitGroup& to_add) {
			units.add(to_add);
		}

		void remove(BWAPI::Unit* unit) {
			units.rem(unit);
			under_construction.rem(unit);
		}

		MyUnitGroup get_worker() {
			// If we happen to have an idle worker, return it
			MyUnitGroup workers = units.harvesters();
			if (!workers.isEmpty()) {
				MyUnitGroup r(workers.at(0));
				workers.rem(r);
				return r;
			}

			// Else, take one from HarvestManager
			MyUnitGroup u = hm.getWorker();
			return u;
		}

		bool has_newly_created() {
			return newly_created.size();
		}
		MyUnitGroup get_newly_created() {
			MyUnitGroup temp;
			temp.add(newly_created);
			newly_created.clear();
			return temp;
		}

		void onFrame() {
			// Move any completed units from under_construction to units.
			MyUnitGroup is_completed = under_construction.is_trained();
			if (is_completed.size() > 0) {
				//BWAPI::Broodwar->printf("moving units to trained");
				under_construction.rem(is_completed);
				units.add(is_completed);
				
				newly_created.add(is_completed.attackers());
			}

			// Add any idle workers to HarvestManager
			MyUnitGroup harvesters = units.harvesters();
			if (harvesters.size() > 0) {
				units.rem(harvesters);
				for (int i = 0; i != harvesters.size(); ++i)
					hm.add(harvesters.at(i));
			}
		}
	};
}