#pragma once

#include <vector>
#include "Gamestate.h"
#include "LocationType.h"
#include "PlanItem.h"
#include <boost/archive/text_oarchive.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <boost/serialization/vector.hpp>

namespace scai {
	struct Case {
		Case(Gamestate state, Plan plan, Plan opponentmodel, int frames, int utility): state(state), plan(plan), opponentmodel(opponentmodel), frames(frames), utility(utility), wins(0), loss(0), match_id("NO MATCH"), ratio_updated(false), child("NO CHILD") {}

		Case(): frames(-1), utility(-1), wins(0), loss(0), match_id("NO MATCH"), ratio_updated(false), child("NO CHILD") {}
		Gamestate state;
		Plan plan;
		Plan opponentmodel;
		int frames;
		int utility;
		int wins;
		int loss;
		std::string match_id;

		//To ensure we calculate win/loss correctly, 1 means win, 0 means not part of matches, -1 means loss
		std::map<std::string, int> matches;

		//This is solely for the use in the CBR module when updating win/loss ratio
		bool ratio_updated;

		//This is the match_id of this cases child
		std::string child;


		friend class boost::serialization::access;
		template <class Archive>
		void serialize(Archive & ar, const unsigned int version) {
			ar & state;
			ar & plan;
			ar & opponentmodel;
			ar & frames;
			ar & utility;
			ar & wins;
			ar & loss;
			ar & match_id;
			ar & child;
			ar & ratio_updated;
			ar & matches;
		}
	};
	typedef std::vector<Case> CaseLibrary;
}