#pragma once
#include "Case.h"
#include <vector>
#include "Gamestate.h"
#include "LocationType.h"
#include "PlanItem.h"
#include <boost/archive/text_oarchive.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <boost/serialization/vector.hpp>
#include <cmath>

namespace scai {
	struct CompositeCase {
	private:

		int round(float number) {
			return (int)(floor(number+0.5f));
		}

		//Finds the similarity between two numbers. 
		float divideInts(int first, int second) {
			//To make sure we dont have zero of divider.
			if(second == 0) {
				first++;
				second++;
			}
			float ans = (float)((float)(first) / (float)(second));

			return ans;
		}



	public:
		CompositeCase(Case case1, Case case2) {
			append_case(case1);
			append_case(case2);
		}
		CompositeCase() {}
		//This is the merged case
		Case composite_case;

		CaseLibrary cases;

		friend class boost::serialization::access;
		template <class Archive>
		void serialize(Archive & ar, const unsigned int version) {
			ar & composite_case;
			ar & cases;
		}

		void append_case(Case case1) {
			cases.push_back(case1);
			calc_composite_case();
		}


		void calc_composite_case() {
			if(cases.size() == 0) {
				return;
			} if(cases.size() == 1) {
				composite_case = cases[0];
				return;
			} else {
				Case temp;
				temp.frames = 0;
				temp.utility = 0;

				//Code for merging all the cases in the library.
				//We only want to merge the fields that are relevant for case matching, mostly in the gamestate
				//And yes, this is (SPARTA!!) monkey coding
				int mBases = 0;
				int eBases = 0;

				for(size_t i = 0 ; i < cases.size() ; i++) {
					temp.wins += cases[i].wins;
					temp.loss += cases[i].loss;

					temp.frames += cases[i].frames;
					temp.state.mMinerals += cases[i].state.mMinerals;
					temp.state.mGas += cases[i].state.mGas;

					temp.state.mMineralsRate += cases[i].state.mMineralsRate;
					temp.state.mGasRate += cases[i].state.mGasRate;

					temp.state.mHarvesters += cases[i].state.mHarvesters;
					temp.state.eHarvesters += cases[i].state.eHarvesters;



					for(std::map<int, int>::const_iterator it = cases[i].state.mResearch.begin() ; it != cases[i].state.mResearch.end() ; ++it) {
						temp.state.mResearch[it->first] += it->second;
					}
					for(std::map<int, int>::const_iterator it = cases[i].state.eResearch.begin() ; it != cases[i].state.eResearch.end() ; ++it) {
						temp.state.eResearch[it->first] += it->second;
					}

					for(size_t j = 0 ; j < cases[i].state.mUnits.size() ; j++) {
						bool flag = false;
						//To ensure that we only add one entry of each unit type, to avoid rounding errors and such
						for(size_t k = 0 ; k < temp.state.mUnits.size() ; k++) {
							if(cases[i].state.mUnits[j].type_id == temp.state.mUnits[k].type_id) {
								temp.state.mUnits[k].count += cases[i].state.mUnits[j].count;
								flag = true;
							}
						}
						if(!flag) {
							temp.state.mUnits.push_back(cases[i].state.mUnits[j]);
						}
					}

					for(size_t j = 0 ; j < cases[i].state.eUnits.size() ; j++) {
						bool flag = false;
						//To ensure that we only add one entry of each unit type, to avoid rounding errors and such
						for(size_t k = 0 ; k < temp.state.eUnits.size() ; k++) {
							if(cases[i].state.eUnits[j].type_id == temp.state.eUnits[k].type_id) {
								temp.state.eUnits[k].count += cases[i].state.eUnits[j].count;
								flag = true;
							}
						}
						if(!flag) {
							temp.state.eUnits.push_back(cases[i].state.eUnits[j]);
						}
					}

					for(size_t j = 0 ; j < cases[i].state.mBuildings.size() ; j++) {
						bool flag = false;
						//To ensure that we only add one entry of each unit type, to avoid rounding errors and such
						for(size_t k = 0 ; k < temp.state.mBuildings.size() ; k++) {
							if(cases[i].state.mBuildings[j].type_id == temp.state.mBuildings[k].type_id) {
								temp.state.mBuildings[k].count += cases[i].state.mBuildings[j].count;
								flag = true;
							}
						}
						if(!flag) {
							temp.state.mBuildings.push_back(cases[i].state.mBuildings[j]);
						}
					}

					for(size_t j = 0 ; j < cases[i].state.eBuildings.size() ; j++) {
						bool flag = false;
						//To ensure that we only add one entry of each unit type, to avoid rounding errors and such
						for(size_t k = 0 ; k < temp.state.eBuildings.size() ; k++) {
							if(cases[i].state.eBuildings[j].type_id == temp.state.eBuildings[k].type_id) {
								temp.state.eBuildings[k].count += cases[i].state.eBuildings[j].count;
								flag = true;
							}
						}
						if(!flag) {
							temp.state.eBuildings.push_back(cases[i].state.eBuildings[j]);
						}
					}

					mBases += cases[i].state.mBases.size();
					eBases += cases[i].state.eBases.size();


				}
				//Need to correct all amounts based on the number of cases!
				float amount_of_cases = (float)cases.size();
				int amount_of_cases_int = cases.size();

				temp.frames /= amount_of_cases_int;

				temp.state.mMinerals /= amount_of_cases_int;

				temp.state.mGas /= amount_of_cases_int;

				temp.state.mMineralsRate /= amount_of_cases;

				temp.state.mGasRate /= amount_of_cases;

				temp.state.mHarvesters /= amount_of_cases_int;

				temp.state.eHarvesters /= amount_of_cases_int;

				for(std::map<int, int>::const_iterator it = temp.state.mResearch.begin() ; it != temp.state.mResearch.end() ; ++it) {
					temp.state.mResearch[it->first] = round((float)(it->second) / amount_of_cases);
				}
				for(std::map<int, int>::const_iterator it = temp.state.eResearch.begin() ; it != temp.state.eResearch.end() ; ++it) {
					temp.state.eResearch[it->first] = round((float)(it->second) / amount_of_cases);
				}
				for(size_t j = 0 ; j < temp.state.mUnits.size() ; j++) {
					temp.state.mUnits[j].count = round((float)(temp.state.mUnits[j].count)/amount_of_cases);
				}

				for(size_t j = 0 ; j < temp.state.eUnits.size() ; j++) {
					temp.state.eUnits[j].count = round((float)(temp.state.eUnits[j].count)/amount_of_cases);
				}

				for(size_t j = 0 ; j < temp.state.mBuildings.size() ; j++) {
					temp.state.mBuildings[j].count = round((float)(temp.state.mBuildings[j].count)/amount_of_cases);
				}

				for(size_t j = 0 ; j < temp.state.eBuildings.size() ; j++) {
					temp.state.eBuildings[j].count = round((float)(temp.state.eBuildings[j].count)/amount_of_cases);
				}

				mBases = round((float)(mBases)/amount_of_cases);
				eBases = round((float)(eBases)/amount_of_cases);


				Gamestate::Bases base;
				base.mine = true;
				base.pos_x = 0;
				base.pos_y = 0;
				base.type = scai::Default;
				//TODO: If we ever change casematcher to use anything else than bases.size this could be troublesome
				for(int j = 0 ; j < mBases ; j++) {
					temp.state.mBases.push_back(base);
				}
				for(int j = 0 ; j < eBases ; j++) {
					temp.state.eBases.push_back(base);
				}


				composite_case = temp;
			}
		}



	};
	typedef std::vector<CompositeCase> CompositeCaseLibrary;

}