#pragma once

#include "Case.h"
#include "CompositeCase.h"
#include "Gamestate.h"
#include "util.h"

#include <fstream>
#include <vector>
#include <string>
#include <algorithm>

#include <boost/archive/text_oarchive.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <boost/archive/binary_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>

namespace scai {
	class Database {
	public:
		Database() { 
			
		}

		Database(const std::string& dbindex_file) {
			std::ifstream ifs(dbindex_file.c_str());
			read_database(dbindex_file);
		}

		size_t Min(size_t a, size_t b) {
			return (a < b) ? a : b;
		}

		void print() {
			std::cout << "Database:" << std::endl;
			std::cout << " cases: " << lib.size() << std::endl;
			std::cout << " clusters: " << lib_comp.size() << std::endl;
			std::cout << " roots:" << root.size() << std::endl;
		}

		void resize(size_t new_size) {
			/* This one shouldn't be a problem, if only clustering worked properly... */
			if (lib_comp.size() != 0)
				throw std::exception("Tried to resize database which has already been clustered");
			if (new_size >= lib.size())
				throw std::exception("Tried to resize database to greater than source");
			
			std::random_shuffle(lib.begin(), lib.end());
			lib.erase(lib.begin() + new_size, lib.end());
		}

		size_t size() {
			return lib.size() + lib_comp.size();
		}

		void join(const Database& other) {
			std::copy(other.lib.begin(), other.lib.end(), std::back_inserter(lib));
			std::copy(other.lib_comp.begin(), other.lib_comp.end(), std::back_inserter(lib_comp));
			std::copy(other.root.begin(), other.root.end(), std::back_inserter(root));
		}

		void append(const CaseLibrary& cases) {
			if(cases.size() > 0) {
				root.push_back(cases[0].match_id);
			}
			std::copy (cases.begin(), cases.end(), std::back_inserter(lib));
		}

		void read_database(const std::string& dbindex) {
			std::ifstream s(dbindex.c_str());
			if (s) {
				boost::archive::text_iarchive ar(s);
				ar >> *this;
			} else {
				std::cerr << "Could not read from: " << dbindex << std::endl;
				return;
			}
		}
		
		void write_database(const std::string& dbindex) {
			std::ofstream s(dbindex.c_str());
			if (s) {
				boost::archive::text_oarchive ar(s);
				ar << *this;
			} else {
				std::cerr << "Could not open db: " << dbindex << std::endl;
				return;
			}
		}
/*
		CaseLibrary get_lib() { return lib; }
		CompositeCaseLibrary get_comp_lib() {return lib_comp;}

		void set_lib(CaseLibrary lib_temp) {
			lib = lib_temp;
		}
		void set_lib_comp(CompositeCaseLibrary lib_comp_temp) {
			lib_comp = lib_comp_temp;
		}
*/
		std::vector<std::string> get_root() {return root;}
		
		
		friend class boost::serialization::access;
		template <class Archive>
		void serialize(Archive & ar, const unsigned int version) {
			ar & lib & lib_comp & root;
		}
		
	
		
		CaseLibrary lib;
		CompositeCaseLibrary lib_comp;
	private:
		//holds all top level parent nodes for casematching
		std::vector<std::string> root;
	};
}