#pragma once
#include <BWAPI.h>
#include <stdio.h>

#include <fstream>

#include <stdarg.h>
#include <sys/stat.h>
#include "Case.h"
#include <boost/foreach.hpp>


namespace scai {
	class DatabaseWriter {
	public:
		/**
		 * Method to write to database. Format is as follows:
		 * First a single integer n which shows how many cases we have
		 *
		 * Then follows n Case descriptions which are structured in this manner, each entry is a line in the txt file:
		 * General information
		 ** One integer, utility of case
		 ** One integer, amount of frames

		 * Gamestate information:
		 ** The map we are player - String
		 ** Player and enemy race
		 ** Minerals and gas, and their respective rate
		 ** Amount of harvesters for us and enemy
		 
		 ** One integer, amount of research
		 ** Our research
		 
		 ** One integer, amount of enemy research
		 ** enemy research

		 ** One integer, amount of units
		 ** our units

		 ** One integer, amount of enemy units
		 ** enemy units

		 ** One integer, amount of buildings
		 ** our buildings

		 ** One integer, amount of enemy buildings
		 ** enemy buildings

		 ** One integer, amount of bases
		 ** our bases

		 ** One integer, amount of enemy bases
		 ** enemy bases

		 * Action list

		 * Opponent model
		 */
		void writeDatabase(CaseLibrary lib) {
			FILE* outfile;
			//NOTE! This action will erase the database file, needed for consistency atm
			int err = fopen_s(&outfile, "bwapi-data\\logs\\databaseTest.txt", "w");
			if(outfile != NULL) {
				//Amount of cases
				writeToFile(outfile, "%d\n", lib.size());

				//For each case
				for(size_t i = 0 ; i < lib.size() ; i++) {
					Case temp = lib.at(i);
					//write utility
					writeToFile(outfile, "%d\n", temp.utility);

					//write frames
					writeToFile(outfile, "%d\n", temp.frames);

					//write map
					writeToFile(outfile, "%s\n", temp.state.map.c_str());

					//write ours and enemy race
					writeToFile(outfile, "%d %d\n", temp.state.my_race_id, temp.state.their_race_id);
					
					//write minerals, gas and rates
					writeToFile(outfile, "%d %d %f %f\n", temp.state.mMinerals, temp.state.mGas, temp.state.mMineralsRate, temp.state.mGasRate);

					// write amount of harvesters, ours and enemy
					writeToFile(outfile, "%d %d\n", temp.state.mHarvesters, temp.state.eHarvesters);

					
					//one integer with amount of research
					writeToFile(outfile, "%d\n", temp.state.mResearch.size());

					//then follows every research we have
					
					for(std::map<int, int>::const_iterator it = temp.state.mResearch.begin() ; it != temp.state.mResearch.end() ; ++it) {
						writeToFile(outfile, "%d %d\n", it->first, it->second);
					}

					//one integer with amount of research
					writeToFile(outfile, "%d\n", temp.state.eResearch.size());

					//then follows every research we have
					
					for(std::map<int, int>::const_iterator it = temp.state.eResearch.begin() ; it != temp.state.eResearch.end() ; ++it) {
						writeToFile(outfile, "%d %d\n", it->first, it->second);
					}


					std::vector<scai::Gamestate::Units> mUnits = temp.state.mUnits;
					//amount of units
					writeToFile(outfile, "%d\n", mUnits.size());
					//Our units
					for(size_t i = 0 ; i < mUnits.size() ; i++) {
						writeToFile(outfile, "%s\n", mUnits[i].t.getName().c_str());
						writeToFile(outfile, "%d\n", mUnits[i].count);
					}

					std::vector<scai::Gamestate::Units> eUnits = temp.state.eUnits;
					//amount of enemy units
					writeToFile(outfile, "%d\n", eUnits.size());
					//enemy units
					for(size_t i = 0 ; i < eUnits.size() ; i++) {
						writeToFile(outfile, "%s\n", eUnits[i].t.getName().c_str());
						writeToFile(outfile, "%d\n", eUnits[i].count);
					}

					
					std::vector<scai::Gamestate::Units> mBuildings = temp.state.mBuildings;
					//amount of buildings
					writeToFile(outfile, "%d\n", mBuildings.size());
					//our buildings
					for(size_t i = 0 ; i < mBuildings.size() ; i++) {
						writeToFile(outfile, "%s\n", mBuildings[i].t.getName().c_str());
						writeToFile(outfile, "%d\n", mBuildings[i].count);
					}

					std::vector<scai::Gamestate::Units> eBuildings = temp.state.eBuildings;
					//amount of enemy buildings
					writeToFile(outfile, "%d\n", eBuildings.size());
					//enemy buildings
					for(size_t i = 0 ; i < eBuildings.size() ; i++) {
						writeToFile(outfile, "%s\n", eBuildings[i].t.getName().c_str());
						writeToFile(outfile, "%d\n", eBuildings[i].count);
					}

					//amount of bases
					writeToFile(outfile, "%d\n", temp.state.mBases.size());
					//our bases
					for(size_t i = 0 ; i < temp.state.mBases.size() ; i++) {
						writeToFile(outfile, "%d\n", temp.state.mBases[i].type);
					}

					//amount of enemy bases
					writeToFile(outfile, "%d\n", temp.state.eBases.size());
					//enemy bases
					for(size_t i = 0 ; i < temp.state.eBases.size() ; i++) {
						writeToFile(outfile, "%d\n", temp.state.eBases[i].type);
					}

					//write amount of actions
					writeToFile(outfile, "%d\n", temp.plan.size());
					//Write each action
					for(size_t i = 0 ; i < temp.plan.size() ; i++) {
						PlanItem temp2 = temp.plan[i];
						switch(temp2.myType) {
						case scai::Build:
							writeToFile(outfile, "BUILD\n");
							writeToFile(outfile, "%s\n", temp2.build.building.getName().c_str());
							break;
						case scai::Train:
							writeToFile(outfile, "TRAIN\n");
							writeToFile(outfile, "%s\n", temp2.train.unit.getName().c_str());
							writeToFile(outfile, "%d\n", temp2.train.count);
							break;
						case scai::Upgrade:
							writeToFile(outfile, "UPGRADE\n");
							writeToFile(outfile, "%s\n", temp2.upgrade.upgrade.getName().c_str());
							break;
						case scai::Attack:
							//TODO: dont use position in attack planitems
							writeToFile(outfile, "ATTACK\n");
							writeToFile(outfile, "%d\n", temp2.attack.location);
							break;
						case scai::Defend:
							//TODO: dont use position in defend planitems
							writeToFile(outfile, "DEFEND\n");
							writeToFile(outfile, "%d\n", temp2.defend.location);
							break;
						case scai::Expand:
							//TODO: Dont use position in expand planitems
							writeToFile(outfile, "EXPAND\n");
							writeToFile(outfile, "%d\n", temp2.expand.location);
							break;
						case scai::Scout:
							//TODO: Dont use position in scout planitems
							writeToFile(outfile, "SCOUT\n");
							writeToFile(outfile, "%d\n", temp2.scout.location);
							break;
						default:
							break;
						}
					}

					//write amount of enemy actions
					writeToFile(outfile, "%d\n", temp.opponentmodel.size());
					//Write each enemy action
					for(size_t i = 0 ; i < temp.opponentmodel.size() ; i++) {
						PlanItem temp2 = temp.opponentmodel[i];
						switch(temp2.myType) {
						case scai::Build:
							writeToFile(outfile, "BUILD\n");
							writeToFile(outfile, "%s\n", temp2.build.building.getName().c_str());
							break;
						case scai::Train:
							writeToFile(outfile, "TRAIN\n");
							writeToFile(outfile, "%s\n", temp2.train.unit.getName().c_str());
							writeToFile(outfile, "%d\n", temp2.train.count);
							break;
						case scai::Upgrade:
							writeToFile(outfile, "UPGRADE\n");
							writeToFile(outfile, "%s\n", temp2.upgrade.upgrade.getName().c_str());
							break;
						case scai::Attack:
							//TODO: dont use position in attack planitems
							writeToFile(outfile, "ATTACK\n");
							writeToFile(outfile, "%d\n", temp2.attack.location);
							break;
						case scai::Defend:
							//TODO: dont use position in defend planitems
							writeToFile(outfile, "DEFEND\n");
							writeToFile(outfile, "%d\n", temp2.defend.location);
							break;
						case scai::Expand:
							//TODO: Dont use position in expand planitems
							writeToFile(outfile, "EXPAND\n");
							writeToFile(outfile, "%d\n", temp2.expand.location);
							break;
						case scai::Scout:
							//TODO: Dont use position in scout planitems
							writeToFile(outfile, "SCOUT\n");
							writeToFile(outfile, "%d\n", temp2.scout.location);
							break;
						default:
							break;
						}
					}

				}
				fclose(outfile);
			} else {
				// TODO write some errors.
			}
		}

	private:
		void writeToFile(FILE* outfile, const char* text, ...) {
			const int BUFFER_SIZE = 1024;
			char buffer[BUFFER_SIZE];

			va_list ap;
			va_start(ap, text);
			vsnprintf_s(buffer, BUFFER_SIZE, BUFFER_SIZE, text, ap);
			va_end(ap);

			fprintf_s(outfile, buffer);
		}
	};
}