#pragma once

#include <vector>
#include <map>

#include <boost/archive/text_oarchive.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <boost/serialization/vector.hpp>
#include <boost/serialization/map.hpp>

#include <boost/foreach.hpp>
#include "log.h"
#include "LocationType.h"

namespace scai{
	struct Gamestate {
		Gamestate():map("NO MAP"), frame(0), my_race_id(0), their_race_id(0), mMinerals(0), mGas(0), mMineralsRate(0.0f), mGasRate(0.0f), mHarvesters(0), eHarvesters(0) {}
		std::string map;

		int frame;

		int my_race_id, their_race_id;

		int mMinerals,  mGas;
		float mMineralsRate, mGasRate;
		
		int mHarvesters, eHarvesters;

		// UpgradeType.id -> level
		std::map<int, int> mResearch;
		std::map<int, int> eResearch;

		struct Units {
			bool mine;
			int count;
			int type_id;
			int pos_x, pos_y;

			bool operator == (const Units& o) {
				return mine == o.mine &&
					count == o.count &&
					type_id == o.type_id &&
					pos_x == o.pos_x && pos_y == o.pos_y;
			}

			friend class boost::serialization::access;
			template <class Archive>
			void serialize(Archive & ar, const unsigned int version) {
				ar & mine;
				ar & count;
				ar & type_id;
				ar & pos_x & pos_y;
			}
		};

		std::vector<Units> mUnits;
		std::vector<Units> mBuildings;

		std::vector<Units> eUnits;
		std::vector<Units> eBuildings;
		

		struct Bases {
			//TODO: If the bool mine is uninitialized we get a segfault when using serialize, rather annoying
			Bases():mine(false){}
			bool mine;
			int pos_x, pos_y;
			scai::LocationType type;

			friend class boost::serialization::access;
			template <class Archive>
			void serialize(Archive & ar, const unsigned int version) {
				ar & mine;
				ar & pos_x & pos_y;
				ar & type;
			}
		};
		std::vector<Bases> mBases;
		std::vector<Bases> eBases;

		// Left out for now.
		// int eMinerals, eMineralsRate, eGas, eGasRate;

		friend class boost::serialization::access;
		template <class Archive>
		void serialize(Archive & ar, const unsigned int version) {
			ar & map;
			ar & frame;
			ar & my_race_id & their_race_id;
			ar & mMinerals & mGas & mMineralsRate & mGasRate;
			ar & mHarvesters & eHarvesters;
			ar & mResearch & eResearch;
			ar & mUnits & mBuildings;
			ar & eUnits & eBuildings;
			ar & mBases & eBases;
		}
	};
}
