#pragma once

#include <string>
#include <sstream>
#include <vector>

#include <boost/archive/text_oarchive.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <boost/serialization/vector.hpp>

#include "LocationType.h"

namespace scai {
	enum PlanItemType {Build, Train, Upgrade, Attack, Defend, Expand, Scout};
	
	struct PlanItem {
		PlanItem(): behaviourId(-1), frameEstimate(0), dependency(-1) {}

		//TODO: replace this with something more sophisticated
		static LocationType getType(int type){
			if(type == 0) 
				return Main;
			if(type == 1)
				return Expansion;
			if(type == 2)
				return Army;
			if(type == 3)
				return Default;
		}

		bool operator == (const PlanItem& other) {
			if (other.myType != myType) return false;
			switch (myType) {
			case scai::Build: return build == other.build;
			case scai::Train: return train == other.train;
			case scai::Upgrade: return upgrade == other.upgrade;
			case scai::Attack: return attack == other.attack;
			case scai::Defend: return defend == other.defend;
			case scai::Expand: return expand == other.expand;
			case scai::Scout: return scout == other.scout;
			default: return false;
			}
		}

		std::string name() {
			switch (myType) {
			case scai::Build: return build.name();
			case scai::Train: return train.name();
			case scai::Upgrade: return upgrade.name();
			case scai::Attack: return attack.name();
			case scai::Defend: return defend.name();
			case scai::Expand: return expand.name();
			case scai::Scout: return scout.name();
			default: return "JAN IS A FAILBOT!";
			}
		}

		
		
		PlanItemType myType;

		struct Build {
			// For build
			int type_id;

			bool operator == (const Build& other) {
				return type_id == other.type_id;
			}

			std::string name() {
				std::stringstream ss;
				ss << "Build";
				return ss.str();
			}

			friend class boost::serialization::access;
			template <class Archive>
			void serialize(Archive & ar, const unsigned int version) {
				ar & type_id;
			}
		};

		struct Train {
			// For train
			int type_id;
			int count;

			bool operator == (const Train& other) {
				return type_id == other.type_id && count == other.count;
			}

			std::string name() {
				std::stringstream ss;
				ss << "Train";
				return ss.str();
			}

			friend class boost::serialization::access;
			template <class Archive>
			void serialize(Archive & ar, const unsigned int version) {
				ar & type_id;
				ar & count;
			}
		};

		struct Upgrade {
			// For upgrade
			int upgrade_id;

			bool operator == (const Upgrade& other) {
				return upgrade_id == other.upgrade_id;
			}

			std::string name() {
				std::stringstream ss;
				ss << "Upgrade";
				return ss.str();
			}

			friend class boost::serialization::access;
			template <class Archive>
			void serialize(Archive & ar, const unsigned int version) {
				ar & upgrade_id;
			}
		};

		struct Attack {
			LocationType location;
			
			Attack(): location(Default) {}

			bool operator == (const Attack& other) {
				return location == other.location;
			}

			std::string name() {
				std::stringstream ss;
				ss << "Attack";
				return ss.str();
			}

			friend class boost::serialization::access;
			template <class Archive>
			void serialize(Archive & ar, const unsigned int version) {
				ar & location;
			}
		};

		struct Defend {
			LocationType location;
			
			Defend(): location(Default) {}

			bool operator == (const Defend& other) {
				return location == other.location;
			}

			std::string name() {
				std::stringstream ss;
				ss << "Defend";
				return ss.str();
			}

			friend class boost::serialization::access;
			template <class Archive>
			void serialize(Archive & ar, const unsigned int version) {
				ar & location;
			}
		};

		struct Expand {
			LocationType location;
			
			Expand(): location(Default) {}

			bool operator == (const Expand& other) {
				return location == other.location;
			}

			std::string name() {
				std::stringstream ss;
				ss << "Expand";
				return ss.str();
			}

			friend class boost::serialization::access;
			template <class Archive>
			void serialize(Archive & ar, const unsigned int version) {
				ar & location;
			}
		};

		struct Scout {
			LocationType location;
			Scout(): location(Default) {}

			bool operator == (const Scout& other) {
				return location == other.location;
			}

			std::string name() {
				std::stringstream ss;
				ss << "Scout";
				return ss.str();
			}

			friend class boost::serialization::access;
			template <class Archive>
			void serialize(Archive & ar, const unsigned int version) {
				ar & location;
			}
		};

		Build build;
		Train train;
		Upgrade upgrade;
		Attack attack;
		Defend defend;
		Expand expand;
		Scout scout;

		//Use this field to see if this planitem has assigned a behaviour to it. If value is -1, no behaviour
		int behaviourId;

		int frameEstimate;

		int dependency;

		friend class boost::serialization::access;
		template <class Archive>
		void serialize(Archive & ar, const unsigned int version) {
			ar & myType;
			ar & build;
			ar & train;
			ar & upgrade;
			ar & attack;
			ar & defend;
			ar & expand;
			ar & scout;
			ar & behaviourId & frameEstimate & dependency;
		}
	};

	typedef std::vector<PlanItem> Plan;
}