#pragma once

#include <vector>
#include <string>
#include <iostream>
#include <fstream>
#include <time.h>

#include <boost/archive/text_oarchive.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <boost/serialization/vector.hpp>

#include "util.h"
#include "Case.h"
#include "Gamestate.h"

namespace scai {
	/**
	 *  This class reads and writes bwapi.ini.
	 *
	 *  Each member represents a value in the ini.
	 *  There is _no_ validation of values
	 */
	struct BwapiIni {
		/* section ai */
		std::string ai;
		std::string ai_dbg;
		std::string tournament;

		/* section auto_menu */
		std::string auto_menu;
		std::string pause_dbg;
		std::string lan_mode;
		std::string auto_restart;
		std::string map;
		std::string mapiteration;
		std::string race;
		std::string enemy_count;
		std::string enemy_race;
		std::string enemy_race_1;
		std::string enemy_race_2;
		std::string enemy_race_3;
		std::string enemy_race_4;
		std::string enemy_race_5;
		std::string enemy_race_6;
		std::string enemy_race_7;
		std::string game_type;
		std::string save_replay;
		std::string wait_for_min_players;
		std::string wait_for_max_players;
		std::string wait_for_time;

		/* section config */
		std::string holiday;

		/* section window */
		std::string windowed;
		std::string left;
		std::string top;
		std::string width;
		std::string height;

		/* section starcraft */
		std::string sound;
		std::string screenshots;

		/* section paths */
		std::string log_path;

		void save(const std::string& filename) {
			std::ofstream ofs(filename.c_str());
			if (ofs) {
				ofs << "[ai]\n";
				ofs << "ai" << " = " << ai << '\n';
				ofs << "ai_dbg" << " = " << ai_dbg << '\n';
				ofs << "tournament" << " = " << tournament << '\n';

				ofs << "[auto_menu]\n";
				ofs << "auto_menu" << " = " << auto_menu << '\n';
				ofs << "pause_dbg" << " = " << pause_dbg << '\n';
				ofs << "lan_mode" << " = " << lan_mode << '\n';
				ofs << "auto_restart" << " = " << auto_restart << '\n';
				ofs << "map" << " = " << map << '\n';
				ofs << "mapiteration" << " = " << mapiteration << '\n';
				ofs << "race" << " = " << race << '\n';
				ofs << "enemy_count" << " = " << enemy_count << '\n';
				ofs << "enemy_race" << " = " << enemy_race << '\n';
				ofs << "enemy_race_1" << " = " << enemy_race_1 << '\n';
				ofs << "enemy_race_2" << " = " << enemy_race_2 << '\n';
				ofs << "enemy_race_3" << " = " << enemy_race_3 << '\n';
				ofs << "enemy_race_4" << " = " << enemy_race_4 << '\n';
				ofs << "enemy_race_5" << " = " << enemy_race_5 << '\n';
				ofs << "enemy_race_6" << " = " << enemy_race_6 << '\n';
				ofs << "enemy_race_7" << " = " << enemy_race_7 << '\n';
				ofs << "game_type" << " = " << game_type << '\n';
				ofs << "save_replay" << " = " << save_replay << '\n';
				ofs << "wait_for_min_players" << " = " << wait_for_min_players << '\n';
				ofs << "wait_for_max_players" << " = " << wait_for_max_players << '\n';
				ofs << "wait_for_time" << " = " << wait_for_time << '\n';

				ofs << "[config]\n";
				ofs << "holiday" << " = " << holiday << '\n';

				ofs << "[window]\n";
				ofs << "windowed" << " = " << windowed << '\n';
				ofs << "left" << " = " << left << '\n';
				ofs << "top" << " = " << top << '\n';
				ofs << "width" << " = " << width << '\n';
				ofs << "height" << " = " << height << '\n';

				ofs << "[starcraft]\n";
				ofs << "sound" << " = " << sound << '\n';
				ofs << "screenshots" << " = " << screenshots << '\n';

				ofs << "[paths]\n";
				ofs << "log_path" << " = " << log_path << std::endl;

				ofs.close();
			} else {
				throw std::exception("Could not write file in ini::save()");
			}
		}

		void load(const std::string& filename) {
			std::ifstream ifs(filename.c_str());
			if (ifs) {
				std::string line;
				while (std::getline(ifs, line)) {
					if (line.size() == 0) continue; /* Empty line */
					else if (line[0] == ';') continue; /* Comment */
					else if (line[0] == '[') continue; /* Section */
					else set_value(line);
				}
			} else {
				/* File does not exist */
			}
		}

		void set_value(std::string line) {
			std::string val = util::split(line, '=');

			std::string key = util::strip(line);
			val = util::strip(val);

			if (key == "ai") ai = val;
			else if (key == "ai_dbg") ai_dbg = val;
			else if (key == "tournament") tournament = val;
			else if (key == "auto_menu") auto_menu = val;
			else if (key == "pause_dbg") pause_dbg = val;
			else if (key == "lan_mode") lan_mode = val;
			else if (key == "auto_restart") auto_restart = val;
			else if (key == "map") map = val;
			else if (key == "mapiteration") mapiteration = val;
			else if (key == "race") race = val;
			else if (key == "enemy_count") enemy_count = val;
			else if (key == "enemy_race") enemy_race = val;
			else if (key == "enemy_race_1") enemy_race_1 = val;
			else if (key == "enemy_race_2") enemy_race_2 = val;
			else if (key == "enemy_race_3") enemy_race_3 = val;
			else if (key == "enemy_race_4") enemy_race_4 = val;
			else if (key == "enemy_race_5") enemy_race_5 = val;
			else if (key == "enemy_race_6") enemy_race_6 = val;
			else if (key == "enemy_race_7") enemy_race_7 = val;
			else if (key == "game_type") game_type = val;
			else if (key == "save_replay") save_replay = val;
			else if (key == "wait_for_min_players") wait_for_min_players = val;
			else if (key == "wait_for_max_players") wait_for_max_players = val;
			else if (key == "wait_for_time") wait_for_time = val;
			else if (key == "holiday") holiday = val;
			else if (key == "windowed") windowed = val;
			else if (key == "left") left = val;
			else if (key == "top") top = val;
			else if (key == "width") width = val;
			else if (key == "height") height = val;
			else if (key == "sound") sound = val;
			else if (key == "screenshots") screenshots = val;
			else if (key == "log_path") log_path = val;
			else {
				std::string err("unknown key in bwapi.ini [" + key + "]");
				throw std::exception(err.c_str());
			}
		}
	};

	/**
	 *  This class parses a *.test -file, and makes the data available to other systems.
	 */
	class TestFileParser {
	public:
		int runs;
		std::string race;
		std::string database;
		std::string map;

		bool parse(const std::string& filename) {
			std::ifstream ifs(filename.c_str());
			if (ifs) {
				std::string ignore; /* Use this for anything we want to ignore */
				
				std::getline(ifs, ignore, ':');
				ifs >> runs;

				std::getline(ifs, ignore, ':');
				ifs >> race;

				std::getline(ifs, ignore, ':');
				ifs >> database;

				std::getline(ifs, ignore, ':');
				std::getline(ifs, map);

				return true;
			} else {
				std::string err = "Could not open file: ";
				err += filename;
				throw std::exception(err.c_str());
			}
		}
	};

	
	/**
	 *  This class tracks games which constitute our "tests", in terms of report results.
	 *
	 *  Each test is stored in a seperate file, which is loaded.  Then, X games are run,
	 *  results collated in a file and reported.
	 */
	class TestRunner {
		friend class TestRunPrinter;

		std::string testfile;
		std::string datafile;

		/* How many games should this test consist of */
		int runs;

		/* Which race should we play against */
		std::string race;

		/* Which database should we use? */
		std::string database;

		/* Which map should we be playing on? */
		std::string map;
		
		std::string get_datafilename(const std::string& testname) {
			util::path_pair p = util::path_split(testname);
			util::file_ext  f = util::file_split(p.second);
			return std::string("results\\") + f.first + ".res";
		}

		/* Are we currently selecting a case? */
		bool started;

		/* When did we start selecting a case? */
		double start_t;

	public:
		/** The possible outcomes of a test-game */
		enum Outcome {Win, Loss, Crash, Unknown};
		
		virtual void start_case() {
			if (!started) {
				started = true;
				start_t = clock();
			} else {
				throw std::exception("Already selecting a case");
			}
		}

		virtual void end_case(const scai::Case& cs, const scai::Gamestate& gs, float match) {
			if (started) {
				started = false;
				double end_t = clock();
				double delta_t = end_t - start_t;

				current.cases.push_back(SelectionData(delta_t, cs.match_id, match));
			} else {
				throw std::exception("Not currently selecting a case");
			}
		}

		virtual std::string get_db() {
			return database;
		}

		/**
		 *  This loads a test-file, and then loads the tracking-data of that testfile.
		 */
		virtual void load_test(const std::string& filename) {
			testfile = filename;
			datafile = get_datafilename(filename);

			TestFileParser parser;
			parser.parse(filename);

			runs = parser.runs;
			race = parser.race;
			database = parser.database;
			map = parser.map;

			load(datafile);
		}

		/**
		 *  Clears all results from this test.
		 */
		virtual void reset() {
			runs = 0;
			race = "";
			database = "";
			map = "";
			games.clear();
			save(datafile);
		}

		/**
		 *  We call this at the start of a match to ensure test-data is properly initialized.
		 */
		virtual void start(const std::string& map) {
			current.outcome = Unknown;
			current.map = map;
			current.run = games.size();

			started = false;
		}

		/**
		 *  
		 */
		virtual void end(int outcome, const std::string& race) {
			/* Dont set outcome if it is known, we have crashed */
			if (current.outcome == Unknown) current.outcome = outcome;
			current.race = race;
			games.push_back(current);
			save(datafile);
		}

		/**
		 *
		 */
		virtual void crash(const std::string& error) {
			current.outcome = Crash;
			current.error = error;
		}

		/**
		 *  Determine whether this is the last game, according to the test-file.
		 *  If so, stop looping games by rewriting the ini.
		 */
		virtual bool check_stop() {
			if (games.size() >= runs) {
				BwapiIni ini;
				ini.load("bwapi-data\\bwapi.ini");
				ini.auto_menu = "OFF";
				ini.save("bwapi-data\\bwapi.ini");
				return true;
			}
			return false;
		}

	private:

		/**
		 *  Loads the internal tracking data from file
		 */
		void load(const std::string& filename) {
			std::ifstream f(filename.c_str());
			if (f) {
				boost::archive::text_iarchive ar(f);
				ar >> games;
			}
		}

		/**
		 *  Save internal tracking data to file
		 */
		void save(const std::string& filename) {
			std::ofstream f(filename.c_str());
			if (f) {
				boost::archive::text_oarchive ar(f);
				ar << games;
			} else {
				/* could not open file */
			}
		}

	
		/**
		 *  This structure contains information about timing during case selection
		 */
		struct SelectionData {
			SelectionData(): time(0.0), case_id("INVALID"), match(-1.0f) {}
			SelectionData(double time, const std::string& id, float match): time(time), case_id(id), match(match) {}

			double time; /* The time to select a case (in whatever unit clock() uses) */
			std::string case_id; /* The id of the selected case */
			float match;

			friend class boost::serialization::access;
			template <class Archive>
			void serialize(Archive & ar, const unsigned int version) {
				ar & time & case_id & match;
			}
		};


		/**
		 *  This represents the data for one tracked game
		 */
		struct TestData {

			int run; /* Which run was this? */
			std::string map; /* Which map did we play on? */
			std::string race; /* Who did we play against? */
			int outcome; /* The outcome of this game */
			std::string error; /* If outcome == Crash, this records the exception-message */
			std::vector<SelectionData> cases; /* Data about case selections */

			friend class boost::serialization::access;
			template <class Archive>
			void serialize(Archive & ar, const unsigned int version) {
				ar & run & map & race & outcome & error & cases;
			}
		};
	public:
		std::vector<TestData> games;

		
		TestData current; /* This is the game we are currently tracking */
	};

	class MockTestRunner : public TestRunner {
		virtual std::string get_db() { return ""; }
		virtual void load_test(const std::string& filename) {}
		virtual void reset() {}
		virtual void start(const std::string& map) {}
		virtual void end(int outcome, const std::string& race) {}
		virtual void crash(const std::string& error) {}
		virtual bool check_stop() { return false; }
	};
}