#pragma once

#include <vector>
#include <string>
#include <stdlib.h>

namespace util {

	/* Strip leading and trailing whitespace */
	std::string strip(const std::string& str);

	/** 
	 *  Split a string on a given character 
	 *  The original string contains the head, the returned string contains the tail.
	 */
	std::string split(std::string& str, char on);


	typedef std::pair<std::string, std::string> path_pair;
	/**
	 *  Splits a windows pathname into path + filename
	 */
	path_pair path_split(const std::string& pathname);

	typedef std::pair<std::string, std::string> file_ext;
	/**
	 *  Split a filename into 'file'.'extension'
	 *  Filename should _NOT_ be a full pathname
	 */
	file_ext file_split(const std::string& filename);

	/**
	 *  Read an integer from a string.
	 *  No fault checking!
	 */
	int parse(const std::string& in);

	/* Return a random number in the range [0, range) */
	size_t random(size_t range);

	template <typename T>
	T abs(T in);
};