#include "util.h"

#include <algorithm>
#include <sstream>

std::string util::strip(const std::string& str) {
	/* Skip the empty string.. */
	if (str.size() == 0) return "";

	std::string::const_iterator start = str.begin();
	std::string::const_iterator end = str.end() - 1;

	while (start < str.end() && *start == ' ') ++start;
	while (end > start && *end == ' ') --end;

	return std::string(start, end + 1);
}

std::string util::split(std::string& str, char on) {
	std::string::const_iterator s = std::find(str.begin(), str.end(), on);
	if (s != str.end()) {
		std::string ret(s + 1, str.end());
		str = std::string(str.begin(), s);
		return ret;
	}
	return "";
}

util::path_pair util::path_split(const std::string& pathname) {
	size_t pos = pathname.find_last_of('\\');
	if (pos != std::string::npos) {
		/* the pathname is a proper path */
		return std::make_pair<std::string, std::string>(pathname.substr(0, pos), pathname.substr(pos + 1));
	} else {
		/* the pathname is a filename only */
		return std::make_pair<std::string, std::string>("", pathname);
	}
}

util::file_ext util::file_split(const std::string& filename) {
	size_t pos = filename.find_last_of('.');
	if (pos != std::string::npos) {
		/* the filename has an extension */
		return std::make_pair<std::string, std::string>(filename.substr(0, pos), filename.substr(pos + 1));
	} else {
		/* the filename lacks an extension */
		return std::make_pair<std::string, std::string>(filename, "");
	}
}

int util::parse(const std::string& in) {
	std::stringstream stream(in);
	int ret;
	stream >> ret;
	return ret;
}

size_t util::random(size_t range) {
	return rand() % range;
}

template <typename T>
T util::abs(T in) {
	return (in < 0) ? -in : in;
}