#pragma once

#include <BWAPI.h>
#include "IBehaviour.h"
#include "MyUnitGroup.h"

namespace scai {

	struct HarvestBehaviour : public IBehaviour {
		HarvestBehaviour(MyUnitGroup units, IBehaviour* parent): IBehaviour(units, parent) {}

		virtual void onStart() {
			//Might be buggish if we have a scout at second position :):)
			MyUnitGroup minerals = scai::Global::i()->all().minerals().withinRange(units.at(1)->getPosition(), 320);

			for (int i = 0; i != units.units.size(); ++i) {
				int patchId = i%minerals.size();
				BWAPI::Unit* patch = minerals.at(patchId);
				units.at(i)->gather(patch);
			}
		}

		virtual void onFrame() {
			/* Empty, for now */
		}
	};

	/*
	 *  This behaviour moves a group of units to a specified position.
	 *  It succeeds when the entire group has stopped moving.
	 */
	struct MoveToBehaviour : public IBehaviour {
		BWAPI::Position position;
		MoveToBehaviour(MyUnitGroup units, IBehaviour* parent, BWAPI::Position position): IBehaviour(units, parent), position(position) {}

		virtual void onStart() {
			for (int i = 0; i != units.units.size(); ++i) {
				units.units[i]->move(position);
			}
		}

		virtual void onFrame() {
			bool done = true;
			for (int i = 0; i != units.units.size(); ++i) {
				if (units.units[i]->isMoving())
					done = false;
			}
			if (done) success();
		}
	};


	/*
	 *  This behaviour instructs a production building to train a single unit.
	 *  Assumption:  units contains only the production building.
	 *  The behaviour succeeds when the building is no longer producing (for whatever reason).
	 */
	struct TrainBehaviour : public IBehaviour {
		BWAPI::UnitType type;
		TrainBehaviour(MyUnitGroup units, IBehaviour* parent, BWAPI::UnitType type): IBehaviour(units, parent), type(type) {}

		virtual void onStart() {
			BWAPI::Broodwar->printf("Behaviour onStart");
			if (!units.units[0]->train(type))
				failure();
			scai::Global::i()->unCommitMinerals(type.mineralPrice());
			scai::Global::i()->unCommitGas(type.gasPrice());
		}

		virtual void onFrame() {
			if (!units.at(0)->isTraining()) {
				success();
			}
		}
	};


	/*
	 *  Instructs a worker (not zerg drones) to construct a building at the specified location.
	 *  Assumption: units contains only the worker
	 *  Succeeds when worker is no longer constructing.
	 */
	struct BuildBehaviour : public IBehaviour {
		BWAPI::Position position;
		BWAPI::UnitType type;

		int frameStart;
		bool issued;
		BWAPI::Unit* building;

		BuildBehaviour(MyUnitGroup units, IBehaviour* parent, BWAPI::Position position, BWAPI::UnitType type)
			: IBehaviour(units, parent)
			, position(position)
			, type(type)
			, issued(false)
			, building(NULL)
		{}

		virtual void onStart() {
			log("Starting build behaviour");
			frameStart = BWAPI::Broodwar->getFrameCount();
			issued = units.at(0)->build(BWAPI::TilePosition(position), type);
			log("ending onstart build behaviour");
		}

		virtual void onFrame() {
			if (!issued) {
				frameStart = BWAPI::Broodwar->getFrameCount();
				issued = units.at(0)->build(BWAPI::TilePosition(position), type);
				if (issued) {	
					scai::Global::i()->unCommitMinerals(type.mineralPrice());
					scai::Global::i()->unCommitGas(type.gasPrice());
				}
			}

			if (building == NULL) {
				MyUnitGroup g = scai::Global::i()->all().mine().mostRecent(type, frameStart);
				if (!g.isEmpty()) building = g.at(0);
			}

			if (building != NULL && building->isCompleted()) {
				log("build behaviour success! behaviour id is: %d", this->behaviourId);
				success();
			}
		}

	};
}