#include "BasicAIModule.h"
#include "log.h"
#include <windows.h>
#include "MyUnitGroup.h"

using namespace BWAPI;

DWORD WINAPI AnalyzeThread(bool & analyzed);

void BasicAIModule::onStart()
{
	//scai::log("This is a crappy test! %s", scai::BehaviourStatus);
	scai::log("\n-------------- NEW GAME ---------------\n");

	//BWAPI::Broodwar->setLocalSpeed(0);
	scai::Global::i();
	this->buildManager = new BuildManager(&this->arbitrator);


	if (Broodwar->isReplay()) return;
	// Enable some cheat flags
	Broodwar->enableFlag(Flag::UserInput);
	//Broodwar->enableFlag(Flag::CompleteMapInformation);

	BWTA::readMap();
	Broodwar->printf("BEGIN: Analyzing map... this may take a minute");
	CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)AnalyzeThread, &analyzed, 0, NULL);
	//BWTA::analyze();
	

	BWAPI::Race race = Broodwar->self()->getRace();
	BWAPI::Race enemyRace = Broodwar->enemy()->getRace();
	BWAPI::UnitType workerType=race.getWorker();


	planner = new scai::Planner(comm);
	executive = new scai::Executive(comm);
	reactive = new scai::Reactive(comm, this->buildManager);

	comm.addLayer(planner);
	comm.addLayer(executive);
	comm.addLayer(reactive);
}

BasicAIModule::~BasicAIModule()
{
	delete this->buildManager;
}
void BasicAIModule::onEnd(bool isWinner)
{
	scai::log("onEnd(%d)",isWinner);
}
void BasicAIModule::onFrame()
{
	
	this->buildManager->update();
	this->arbitrator.update();
	try {
		//scai::log("pre reactive");
		reactive->onFrame();
		//scai::log("pre exec");
		executive->onFrame();
		//scai::log("pre plan");
		planner->onFrame();
		//scai::log("frame end");
	} catch (std::exception e) {
		Broodwar->printf("ERROR: %s", e.what());
	} catch (...) {
		Broodwar->printf("ERROR: Unknown exception");
	}


	/*
	scai::MyUnitGroup all = scai::Global::i()->all().mine();
	for (int i = 0; i != all.size(); ++i)
		Broodwar->drawTextScreen(5, 5 + 10*i, "%d: %s", i, all.at(i)->getType().getName().c_str());
	*/
}

void BasicAIModule::onUnitCreate(BWAPI::Unit* unit) {
	if (Broodwar->isReplay()) return;

	int frame = BWAPI::Broodwar->getFrameCount();
	unit->setClientInfo(new scai::UnitData(frame));
	executive->onUnitCreate(unit);
}

void BasicAIModule::onUnitDestroy(BWAPI::Unit* unit)
{
	if (Broodwar->isReplay()) return;

	delete unit->getClientInfo();

	this->buildManager->onRemoveUnit(unit);
	this->arbitrator.onRemoveObject(unit);
}

void BasicAIModule::onUnitDiscover(BWAPI::Unit* unit)
{
	if (Broodwar->isReplay()) return;
}
void BasicAIModule::onUnitEvade(BWAPI::Unit* unit)
{
	if (Broodwar->isReplay()) return;
}

void BasicAIModule::onUnitMorph(BWAPI::Unit* unit)
{
	if (Broodwar->isReplay()) return;
}
void BasicAIModule::onUnitRenegade(BWAPI::Unit* unit)
{
	if (Broodwar->isReplay()) return;
}

void BasicAIModule::onSendText(std::string text)
{	
	Broodwar->sendText("%s",text.c_str());
}

DWORD WINAPI AnalyzeThread(bool & analyzed)
{
	BWTA::analyze();
	Broodwar->printf("END:   Analyzing map... this may take a minute");
	analyzed = true;
	return 0;
}
