#pragma once

#include "log.h"

#include "ILayer.h"
#include "PlanItem.h"

namespace scai {
	class Executive : public ILayer {
		bool recieved;
		Communication& comm;

		bool harvesting;

		Plan current;

		bool canTrain(BWAPI::UnitType unit) {
			scai::Global* g = scai::Global::i();
			return ((g->getAvailableMinerals() >= unit.mineralPrice()) &&
					(g->getAvailableGas() >= unit.gasPrice()) &&
					(!g->all().producers(unit).not_training().isEmpty()));
		}

		bool canBuild(BWAPI::UnitType unit) {
			scai::Global* g = scai::Global::i();
			return ((g->getAvailableMinerals() >= unit.mineralPrice()) &&
					(g->getAvailableGas() >= unit.gasPrice()));
		}
	public:

		virtual void onUnitCreate(BWAPI::Unit* unit) {

		}

		virtual void onFrame() {
			while (comm.hasNext(this)) {
				log("Executive has message.");
				Message msg = comm.getNext(this);

				log("Executive got message with id: %d", msg.id);


				if (msg.id == scai::NewPlan) {
					if (current.size()) BWAPI::Broodwar->printf("WARNING: Executive recieved plan while one was still executing");
					try {
						current = boost::any_cast<Plan>(msg.get("plan"));
					} catch (boost::bad_any_cast e) {
						BWAPI::Broodwar->printf("ERROR: Executive recieved message 'NewPlan' without key 'plan'");
					}
				} 

				//TODO: Currently using planstatus for behaviours, change when it works again :), also getString is bugged for me
				else if(msg.id == scai::PlanStatus){
					log("Executive got behaviour update");
					if(current.size()){
						log("checking if behaviour id matches with the one we have in database, values plan %d and msg %d and string %s", current.front().behaviourId, msg.behaviourId, msg.getString("behaviourStatus"));
						if(current.front().behaviourId == msg.behaviourId /*&& msg.getString("behaviourStatus") == "success"*/) {
							log("behaviour was success, erase it!");
							current.erase(current.begin());
						}/*
						else if(current.front().behaviourId == msg.behaviourId && msg.getString("behaviourStatus") == "failure") {
							log("behaviour was failure, erase it!");
							current.erase(current.begin());
						}*/
					}
				}
				
				else {
					BWAPI::Broodwar->printf("WARNING: Executive recieved unknown message with id: %s", msg.id);
				}
			}

			for (int i = 0; i != current.size(); ++i)
				BWAPI::Broodwar->drawTextScreen(5, 5 + 8 * i, "Planitem: %s", current[i].name());

			
			

			if (!harvesting) {
				Message harvest;
				harvest.id = NewBehaviour;
				harvest.putString("behaviour", "harvest");
				harvest.put("harvesters", Global::i()->all().mine().harvesters());
				comm.sendDn(this, harvest);
				harvesting = true;
			}
			

			/*TODO: Currently one item at the time, fix behaviourId for concurrency*/

			if (current.size()) {
				PlanItem& next = current.front();
				
				switch (next.myType) {
				case scai::Build:
					if (canBuild(next.build.building) && next.behaviourId == -1) {
						log("creating build behaviour");
						next.behaviourId = scai::Global::i()->getNextId();
						log("Id is: %d", next.behaviourId);
						order_build(next.build.building, next.behaviourId);
					}
					break;
				case scai::Train:
					if (canTrain(next.train.unit)) {
						
						BWAPI::Broodwar->printf("INFO: Executive ordered train");
						order_train(next.train.unit);

						if (next.train.count)
							next.train.count--;
						
						if (next.train.count == 0)
							current.erase(current.begin());
					}
					break;
				case scai::Scout:
					BWAPI::Broodwar->printf("WARNING: Unimplemented behaviour SCOUT");
					current.erase(current.begin());
					break;
				default:
					BWAPI::Broodwar->printf("WARNING: Unimplemented behaviour");
					//TODO: Fix so we dont toss the whole damn plan!
					//Remove unhandled planitems so we dont stop
					current.erase(current.begin());
				}
			}
			//plan is empty, tell planning layer and delete it!
			else if(current.size() == 0) {
				//TODO: use a plan id! not necessary right now with dummy impl
				log("INFO: plan empty");
				Message planFinished;
				planFinished.id = PlanFinished;
				planFinished.putString("planFinished", "Plan id?");
				comm.sendUp(this, planFinished);
				log("INFO: requested new plan");
			}
		}

		Executive(Communication & comm): comm(comm), recieved(false), harvesting(false) {
		
		}

		void order_build(BWAPI::UnitType type, int behaviourId) {
			scai::Global::i()->commitMinerals(type.mineralPrice());
			scai::Global::i()->commitGas(type.gasPrice());

			MyUnitGroup unit = scai::Global::i()->all().mine().harvesters()[0];
			
			Message get_builder;
			get_builder.id = ReleaseUnit;
			get_builder.put("unit", unit);
			comm.sendDn(this, get_builder);

			Message build;
			build.id = NewBehaviour;
			build.behaviourId = behaviourId;
			build.putString("behaviour", "build");
			build.put("type", type);
			build.put("worker", unit);
			comm.sendDn(this, build);
		}

		void order_train(BWAPI::UnitType type) {
			scai::Global::i()->commitMinerals(type.mineralPrice());
			scai::Global::i()->commitGas(type.gasPrice());

			MyUnitGroup producer = scai::Global::i()->all().producers(type);

			Message build;
			build.id = NewBehaviour;
			build.behaviourId = scai::Global::i()->getNextId();
			build.putString("behaviour", "train");
			build.put("unit", type);
			build.put("producer", producer);
			comm.sendDn(this, build);
		}
	};
}