#pragma once

#include "BasicAIModule.h"
#include "MyUnitGroup.h"
#include <BWAPI.h>
#include <BuildingPlacer.h>
#include "IdFactory.h"

namespace scai {
	struct Global {
	private:
		Global(): committedMinerals(0), committedGas(0) {		};
		Global(Global const&) {};
		Global& operator = (Global const&) {};
		Gamestate state;
		
		BuildingPlacer placer;
		int committedMinerals;
		int committedGas;
		IdFactory id;
	public:
		
		int getNextId(){
			return id.getNextId();
		}

		Gamestate getGamestate(){
			return state;
		}

		BWAPI::TilePosition getBuildLocationNear(BWAPI::TilePosition pos, BWAPI::UnitType type) {
			return placer.getBuildLocationNear(pos, type);
		}

		BWAPI::TilePosition startLocation() {
			return BWAPI::Broodwar->self()->getStartLocation();
		}

		/**
		 *  Return the amount of minerals which are currently available to use.
		 */
		int getAvailableMinerals() {
			return BWAPI::Broodwar->self()->minerals() - committedMinerals;
		}


		/**
		 *  Return the amount of gas available to you now.
		 */
		int getAvailableGas() {
			return BWAPI::Broodwar->self()->gas() - committedGas;
		}


		/**
		 *  Use this to mark some minerals as spent
		 *  Remember to un-commit them later!
		 */
		void commitMinerals(int minerals) {
			committedMinerals += minerals;
		}


		/**
		 *  Use this to mark minerals as used.
		 */
		void unCommitMinerals(int minerals) {
			committedMinerals -= minerals;
		}


		/**
		 *  Use this to mark some gas as spent
		 *  Remember to un-commit them later!
		 */
		void commitGas(int gas) {
			committedGas += gas;
		}

		/**
		 *  Use this to mark gas as used.
		 */
		void unCommitGas(int gas) {
			committedGas -= gas;
		}


		/**
		 *  Get the current frame of the game
		 */
		int getCurrentFrame() {
			return BWAPI::Broodwar->getFrameCount();
		}


		/**
		 *  Get a group of _all_ units currently in the game.
		 */
		MyUnitGroup all() {
			std::set<BWAPI::Unit*> u = BWAPI::Broodwar->getAllUnits();
			return MyUnitGroup(std::vector<BWAPI::Unit*>(u.begin(), u.end()));
		}

		static Global* i() {
			static Global instance;
			return &instance;
		}
	};
}