#pragma once

#include <string>
#include <map>
#include <boost/shared_ptr.hpp>
#include <boost/any.hpp>
#include "MessageTypes.h"





namespace scai {

	typedef std::map<std::string, boost::any> Payload;
	class Message {
	public:
		Message(): payload(new Payload()) {}

		MessageId id;
		//used by behaviours when being created by reactive
		int behaviourId;
		boost::shared_ptr<Payload> payload;

		bool operator == (const Message& o) {
			return (id == o.id) && (payload == o.payload);
		}

		boost::any get(std::string key) {
			return (*payload)[key];
		}
		void put(std::string key, boost::any val) {
			(*payload)[key] = val;
		}

		template <typename T>
		T getCast(std::string key) {
			return boost::any_cast<T>(get(key));
		}

		std::string getString(std::string key) {
			return boost::any_cast<std::string>(get(key));
		}

		void putString(std::string key, std::string value) {
			put(key, value);
		}
	};
}