#pragma once

#include <exception>
#include <vector>
#include <algorithm>
#include <BWAPI.h>

#include "Predicates.h"

namespace scai {

	struct UnitData {
		UnitData(): age(0), assigned(false) {}
		UnitData(int age): age(age), assigned(false) {}

		int age;
		bool assigned;
	};

	int getAge(BWAPI::Unit* unit);

	/**
	 *  This represents a group of units under control of a behaviour.
	 */
	struct MyUnitGroup {

		void release(const MyUnitGroup& bye) {
			std::vector<BWAPI::Unit*> remove;
			for (int i = 0; i != bye.units.size(); ++i)
				remove.push_back(bye.units[i]);
			for (int i = 0; i != remove.size(); ++i) {
				std::vector<BWAPI::Unit*>::iterator it = std::find(units.begin(), units.end(), remove.at(i));
				if (it != units.end()) units.erase(it);
			}
		}

		bool isEmpty() {
			return units.size() == 0;
		}

		MyUnitGroup unassigned() {
			std::vector<BWAPI::Unit*> other;
			for (int i = 0; i != units.size(); ++i) {
				if (((UnitData*)units[i]->getClientInfo())->assigned == false) other.push_back(units[i]);
			}
			return MyUnitGroup(other);
		}

		MyUnitGroup closest(const BWAPI::Position& pos) {
			int c = 0;
			int min_dist = units[c]->getDistance(pos);

			for (int i = 0; i != units.size(); ++i) {
				int dist = units[i]->getDistance(pos);
				if (dist < min_dist) {
					min_dist = dist;
					c = i;
				}
			}
			return MyUnitGroup(units[c]);
		}


		//return the units that are within a certain range of the position
		MyUnitGroup withinRange(const BWAPI::Position& pos, int range) {
			MyUnitGroup temp;

			for(int i = 0 ; i < units.size() ; i++) {
				int dist = units[i]->getDistance(pos);
				if(dist < range){
					temp.add(units[i]);
				}
			}

			return temp;

		}

		MyUnitGroup mostRecent(const BWAPI::UnitType& type, int noOlderThan=0) {
			for (int i = 0; i != units.size(); ++i) {
				int age = ((UnitData*) units[i]->getClientInfo())->age;
				if (units[i]->getType() == type && age >= noOlderThan)
					return MyUnitGroup(units[i]);
			}
			return MyUnitGroup();
		}

		MyUnitGroup minerals() {
			return by_type(BWAPI::UnitTypes::Resource_Mineral_Field);
		}

		MyUnitGroup by_type(const BWAPI::UnitType& type) {
			std::vector<BWAPI::Unit*> other;
			for (int i = 0; i != units.size(); ++i)
				if (units[i]->getType() == type) other.push_back(units[i]);
			return MyUnitGroup(other);
		}

		MyUnitGroup producers(const BWAPI::UnitType& type) {
			BWAPI::UnitType producer = type.whatBuilds().first;

			return by_type(producer);
		}

		MyUnitGroup not_training() {
			std::vector<BWAPI::Unit*> other;
			for (int i = 0; i != units.size(); ++i)
				if (!units[i]->isTraining()) other.push_back(units[i]);
			return MyUnitGroup(other);
		}

		MyUnitGroup mine() {
			std::vector<BWAPI::Unit*> others;
			for (int i = 0; i != units.size(); ++i)
				if (scai::IsMine(units[i])) others.push_back(units[i]);
			return MyUnitGroup(others);
		}

		MyUnitGroup hostiles() {
			std::vector<BWAPI::Unit*> others;
			for (int i = 0; i != units.size(); ++i)
				if (scai::IsEnemys(units[i])) others.push_back(units[i]);
			return MyUnitGroup(others);
		}

		MyUnitGroup harvesters() {
			std::vector<BWAPI::Unit*> others;
			for (int i = 0; i != units.size(); ++i)
				if (scai::IsHarvester(units[i])) others.push_back(units[i]);
			return MyUnitGroup(others);
		}



		void add(BWAPI::Unit* unit) {
			units.push_back(unit);
		}

		void rem(BWAPI::Unit* unit) {
			units.erase(std::find(units.begin(), units.end(), unit));
		}

		BWAPI::Unit* at(const std::vector<BWAPI::Unit*>::size_type& index) {
			if (index < 0 || index >= units.size()) throw std::exception("Invalid index in MyUnitGroup::at()");
			return units[index];
		}

		std::vector<BWAPI::Unit*>::size_type size() {
			return units.size();
		}



		std::vector<BWAPI::Unit*> units;
		
		MyUnitGroup(): units() {}
		MyUnitGroup(BWAPI::Unit* unit) {
			units.push_back(unit);
		}
		MyUnitGroup(const MyUnitGroup& other): units(other.units) {}
		MyUnitGroup(const std::vector<BWAPI::Unit*>& units): units(units) {}

		MyUnitGroup operator [] (const std::vector<BWAPI::Unit*>::size_type& index) {
			return MyUnitGroup(units[index]);
		}
	};

}