#pragma once
#include "PlanItem.h"
#include "Gamestate.h"
#include <boost/foreach.hpp>
#include <BWAPI.h>

namespace scai {
	class PlanGenerator {
	private: 
		//TODO: temporary for progressing in plans!
		int planCount;

		public:
			PlanGenerator():planCount(0) {}
		/**
		* Currently hardcoded
		* Needs to use gamestate to find gametime and locations in the future in the future
		*/
			void addTrain(Plan& plan, BWAPI::UnitType unitType, int count){
				PlanItem temp;
				temp.myType = scai::Train;
				temp.train.unit = unitType;
				temp.train.count = count;
				plan.push_back(temp);
			}
			void addBuilding(Plan& plan, BWAPI::UnitType unitType){
				PlanItem temp;
				temp.myType = scai::Build;
				temp.build.building = unitType;
				plan.push_back(temp);
			}

			void addAttack(Plan& plan, BWAPI::Position pos){
				PlanItem temp;
				temp.myType = scai::Attack;
				temp.attack.position = pos;
				plan.push_back(temp);
			}

			void addDefend(Plan& plan, BWAPI::Position pos){
				PlanItem temp;
				temp.myType = scai::Defend;
				temp.defend.position = pos;
				plan.push_back(temp);
			}

			void addScout(Plan& plan, BWAPI::Position pos){
				PlanItem temp;
				temp.myType = scai::Scout;
				temp.scout.position = pos;
				plan.push_back(temp);
			}

			void addExpand(Plan& plan, BWAPI::Position pos){
				PlanItem temp;
				temp.myType = scai::Expand;
				temp.expand.position = pos;
				plan.push_back(temp);
			}
			void addUpgrade(Plan& plan, BWAPI::UpgradeType upgrade){
				PlanItem temp;
				temp.myType = scai::Upgrade;
				temp.upgrade.upgrade = upgrade;
				plan.push_back(temp);
			}



			
			virtual Plan generatePlan(Gamestate state){

				if(planCount == 0){
					planCount++;
					Plan initial;

					addTrain(initial, BWAPI::UnitTypes::Protoss_Probe, 4);


					addBuilding(initial, BWAPI::UnitTypes::Protoss_Pylon);

					//currently a dummy position as we dont have any fancy information usage here
					BWAPI::Position temp(0,0);
					addScout(initial, temp);

					addTrain(initial, BWAPI::UnitTypes::Protoss_Probe, 1);


			
					return initial;
				}

				else if(planCount == 1){
					planCount++;
					Plan initialAttack;
					

					addTrain(initialAttack, BWAPI::UnitTypes::Protoss_Probe, 3);
					addBuilding(initialAttack, BWAPI::UnitTypes::Protoss_Gateway);
					addTrain(initialAttack, BWAPI::UnitTypes::Protoss_Zealot, 3);
					addBuilding(initialAttack, BWAPI::UnitTypes::Protoss_Gateway);
					addBuilding(initialAttack, BWAPI::UnitTypes::Protoss_Pylon);
					addTrain(initialAttack, BWAPI::UnitTypes::Protoss_Zealot, 4);

					//Ugly call to BWAPI to find enemy base on 2 player map :)
					BOOST_FOREACH(BWAPI::TilePosition o, BWAPI::Broodwar->getStartLocations()){
						
						if(o.x() != BWAPI::Broodwar->self()->getStartLocation().x()){
							BWAPI::Position temp(o.x() * 32, o.y() * 32);
							addAttack(initialAttack, temp);
							break;
						}
					}

					addBuilding(initialAttack, BWAPI::UnitTypes::Protoss_Pylon);
					addTrain(initialAttack, BWAPI::UnitTypes::Protoss_Probe, 3);

					return initialAttack;
				}

				//still in first 10 minutes
				else if(planCount == 2){
					planCount++;
					Plan midAttack;

					addTrain(midAttack, BWAPI::UnitTypes::Protoss_Probe, 4);
					addBuilding(midAttack, BWAPI::UnitTypes::Protoss_Pylon);
					addBuilding(midAttack, BWAPI::UnitTypes::Protoss_Assimilator);
					addBuilding(midAttack, BWAPI::UnitTypes::Protoss_Cybernetics_Core);
					addTrain(midAttack, BWAPI::UnitTypes::Protoss_Zealot, 4);
					addBuilding(midAttack, BWAPI::UnitTypes::Protoss_Gateway);
					addBuilding(midAttack, BWAPI::UnitTypes::Protoss_Pylon);
					addBuilding(midAttack, BWAPI::UnitTypes::Protoss_Pylon);
					addTrain(midAttack, BWAPI::UnitTypes::Protoss_Dragoon, 5);

					//Ugly call to BWAPI to find enemy base on 2 player map :)
					BOOST_FOREACH(BWAPI::TilePosition o, BWAPI::Broodwar->getStartLocations()){
						
						if(o.x() != BWAPI::Broodwar->self()->getStartLocation().x()){
							BWAPI::Position temp(o.x()*32, o.y()*32);
							addAttack(midAttack, temp);
							break;
						}
					}

					return midAttack;

					
				}
				else{
					Plan repeatAttack;

					addBuilding(repeatAttack, BWAPI::UnitTypes::Protoss_Pylon);
					addTrain(repeatAttack, BWAPI::UnitTypes::Protoss_Zealot, 4);
					addBuilding(repeatAttack, BWAPI::UnitTypes::Protoss_Pylon);
					addTrain(repeatAttack, BWAPI::UnitTypes::Protoss_Dragoon, 5);
					
					//Ugly call to BWAPI to find enemy base on 2 player map :)
					BOOST_FOREACH(BWAPI::TilePosition o, BWAPI::Broodwar->getStartLocations()){
						
						if(o.x() != BWAPI::Broodwar->self()->getStartLocation().x()){
							BWAPI::Position temp(o.x()*32, o.y()*32);
							addAttack(repeatAttack, temp);
							break;
						}
					}
					
					return repeatAttack;
				}
			}

			//dummy implementation
			virtual bool validatePlan(Gamestate state, Plan plan){
				return true;
			}
	

	

	};
}