#pragma once

#include "ILayer.h"
#include "log.h"
#include <BuildManager.h>
#include <BuildingPlacer.h>
#include "IBehaviour.h"
#include "BaseBehaviours.h"


/*
 Note to self:  Crashes on 8 probes.  WHY??
*/

namespace scai {

	struct RootBehaviour : public IBehaviour {
		std::vector<IBehaviour *> children;
		std::vector<IBehaviour *> remove;
		std::vector<Message> behaviourMsg;

		RootBehaviour(): IBehaviour(MyUnitGroup(), NULL) {}

		void addChild(IBehaviour* child) {
			BWAPI::Broodwar->printf("INFO: Adding behaviour with type: %s", child->name());
			for (int i = 0; i != child->units.size(); ++i) {
				UnitData* d = (UnitData*) child->units.at(i)->getClientInfo();
				d->assigned = true;
				child->units.at(i)->setClientInfo(d);
			}
			child->onStart();
			children.push_back(child);
		}

		void remChild(IBehaviour* child) {
			std::vector<IBehaviour*>::iterator i = std::find(children.begin(), children.end(), child);

			if (i == children.end()) {
				log("WARNING: root behaviour could not find the given child");
			} else {
				//delete *i;
				children.erase(i);
			}
		}

		void onFrame() {
			for (int i = 0; i != children.size(); ++i)
				children[i]->onFrame();

			int s = remove.size();
			for (int i = s - 1; i >= 0; --i) {
				remChild(remove[i]);
			}
			remove.clear();
		}

		//TODO: Currently using planstatus for behaviours, change when it works again :)
		void onChildFailure(IBehaviour* child) {
			//Add a message for reactive to send to executive
			if(child->behaviourId != -1){
				Message failure;
				failure.id = scai::PlanStatus;
				failure.behaviourId = child->behaviourId;
				failure.putString("behaviourStatus", "Failure");
				behaviourMsg.push_back(failure);
			}
			remove.push_back(child);
		}


		//TODO: Currently using planstatus for behaviours, change when it works again :)
		void onChildSuccess(IBehaviour* child) {
			//Add a message for reactive to send to executive
			log("child was successful!");
			if(child->behaviourId != -1){
				log("Creating message to be sent for behaviourId %d", child->behaviourId);
				Message success;
				success.id = scai::PlanStatus;
				success.behaviourId = child->behaviourId;
				success.putString("behaviourStatus", "Success");
				behaviourMsg.push_back(success);
			}
			remove.push_back(child);
		}
		
		//TODO: Fix so it doesnt retrieve if there are no messages, currently safely handled in reactive though
		Message retrieveMessage() {
			log("Retrieving message from root");
			Message r = behaviourMsg.front();
			log(" Sending message with info: %d %d", r.id,  r.behaviourId);
			behaviourMsg.erase(std::find(behaviourMsg.begin(), behaviourMsg.end(), r));
			return r;
				

		}
		bool hasMoreMessages() {
			if(behaviourMsg.size()){
				return true;
			}
			else {
				return false;
			}
		}

		
	};
	
	class Reactive : public ILayer {
		Communication & comm;
		//BuildManager* buildManager;
		RootBehaviour root;
		std::vector<Message> msgBuffer;

		
	public:
		
		virtual void onFrame() {
			//log("reactive::onFrame() start");
			while (comm.hasNext(this)) {
				log("while start!");
				Message msg = comm.getNext(this);
				log("msg id %d", msg.id);
				log("while mid!");

				if (msg.id == ReleaseUnit) {
					log ("release start");
					root.releaseUnits(msg.getCast<MyUnitGroup>("unit"));
					log ("release end");
				} else if (msg.getString("behaviour") == "train") {
					spawn_train(msg.getCast<BWAPI::UnitType>("unit"), msg.getCast<MyUnitGroup>("producer"), msg.behaviourId);
				} else if (msg.getString("behaviour") == "harvest") {
					spawn_harvest(msg.getCast<MyUnitGroup>("harvesters"));
				} else if (msg.getString("behaviour") == "build") {
					MyUnitGroup builder = msg.getCast<MyUnitGroup>("worker");
					log("Reactive creating build behaviour with id: %d", msg.behaviourId);
					spawn_build(msg.getCast<BWAPI::UnitType>("type"), builder, msg.behaviourId);
				} else {
					log("WARN: Reactive: Unknown message id");
					BWAPI::Broodwar->printf("WARN: Reactive: Unknown message id");
				}
				log("while end!");
			}
			while(root.hasMoreMessages()) {
				Message r = root.retrieveMessage();
				comm.sendUp(this, r);
			}
			/*
			for (int i = 0; i != root.children.size(); ++i)
				BWAPI::Broodwar->drawTextScreen(5, 5+8*i, "Behaviour: %s", root.children[i]->name());
			*/
			//log("reactive::onFrame() mid");
			root.onFrame();
			//log("reactive::onFrame() end");
		}

		Reactive(Communication & comm, BuildManager* buildManager): comm(comm) {}

		void spawn_build(BWAPI::UnitType type, const MyUnitGroup& builder, int behaviourId) {
			log("INFO: pre spawn_build");
			BWAPI::TilePosition pos = scai::Global::i()->getBuildLocationNear(scai::Global::i()->startLocation(), type);
			
			IBehaviour* build = new scai::BuildBehaviour(builder, &root, BWAPI::Position(pos), type);
			build->behaviourId = behaviourId;
			
			root.addChild(build);
			log("INFO: post add IBehaviour");
		}

		void spawn_train(BWAPI::UnitType type, const MyUnitGroup& producer, int behaviourId) {
			IBehaviour* train = new scai::TrainBehaviour(producer, &root, type);
			root.addChild(train);
		}

		void spawn_harvest(const MyUnitGroup& harvesters) {
			IBehaviour* harvest = new scai::HarvestBehaviour(harvesters, &root);
			root.addChild(harvest);
		}
	};
}