#include "BasicAIModule.h"
#include "log.h"
#include <windows.h>
#include <sstream>

#include <boost/archive/text_oarchive.hpp>
#include <boost/archive/text_iarchive.hpp>

using namespace BWAPI;

DWORD WINAPI AnalyzeThread(bool & analyzed);

const char* tracesr = "traces\\traces";

void BasicAIModule::load_traces()
{
	std::ifstream f("traces\\traces");
	if (!f) return;
	boost::archive::text_iarchive ar(f);
	ar >> traces; 
}

void BasicAIModule::save_traces()
{
	std::ofstream f("traces\\traces");
	boost::archive::text_oarchive ar(f);
	ar << traces;
}

void BasicAIModule::onStart()
{
	do_write = true;
	scai::log("load pre");
	load_traces();
	scai::log("log post");
	if (std::find(traces.begin(), traces.end(), Broodwar->mapPathName()) != traces.end()) {
		scai::log("This replay has already been traced, stopping.");
		scai::log(Broodwar->mapPathName().c_str());
		scai::log("-------------- END TRACE ---------------");
		stop_replays();
		do_write = false;
		Broodwar->leaveGame();
		return;
	}

	int count = 0;
	for(size_t i = 0 ; i < Broodwar->getPlayers().size() ; i++) {
		if(!(Broodwar->getPlayer(i)->isNeutral())) {
			count++;
		}
	}
	if(count != 2) {
		scai::log("More than 2 players, no tracing replay %s, count is %d", Broodwar->mapPathName().c_str(), count);
		stop_replays();
		do_write = false;
		Broodwar->leaveGame();
		return;
	}

	scai::log("onStart");
	scai::log("-------------- NEW TRACE ---------------");
	scai::log(Broodwar->mapPathName().c_str());
	if (Broodwar->isReplay()) {
		Broodwar->setGUI(false);
		Broodwar->setLocalSpeed(0);
		Broodwar->enableFlag(Flag::CompleteMapInformation);
		scai::log("Map is: %s", Broodwar->mapPathName().c_str());

		int i = 0; 
		tracer.onStart();
	}
}

BasicAIModule::~BasicAIModule()
{
	
}
void BasicAIModule::onEnd(bool isWinner)
{
	if (do_write) {

		for(size_t i = 0 ; i < Broodwar->getPlayers().size() ; i++) {
			scai::log("Player at pos %d did win? %d name is %s", i, Broodwar->getPlayer(i)->isVictorious(), Broodwar->getPlayer(i)->getName().c_str());
		}
		
		bool noWinner = true;
		//Find out who the winner is!
		for(size_t i = 0 ; i < Broodwar->getPlayers().size() ; i++) {
			if(!(Broodwar->getPlayer(i)->isNeutral()) && Broodwar->getPlayer(i)->isVictorious()) {
				scai::log("Found a winner at %d with name %s!", i, Broodwar->getPlayer(i)->getName().c_str());
				tracer.onEnd(Broodwar->getPlayer(i));
				noWinner = false;
				break;
			}
		}
		int count = 0;
		for(size_t i = 0 ; i < Broodwar->getPlayers().size() ; i++) {
			if(!(Broodwar->getPlayer(i)->isNeutral())) {
				count++;
			}
		}
		//Should not store the trace if theres no winner (dont ask, it happens some times)
		if(noWinner) {
			scai::log("Could not find winner for game: %s",Broodwar->mapFileName().c_str());
			//some error? 
		} else if(count != 2) {
			//If we have more (or, god forbid, less:P) players than 2 then we cant find store the game properly
			scai::log("Didn't have exactly 2 players, didnt store for game: %s" , Broodwar->mapFileName().c_str());
		} else {
			std::stringstream name;
			name << "traces\\" << Broodwar->mapFileName() << ".trace";
			std::ofstream ofs(name.str().c_str());
			if (ofs) {
				boost::archive::text_oarchive oar(ofs);
				oar << tracer.trace;
				ofs.close();
			}
			traces.push_back(Broodwar->mapPathName());
			save_traces();
		}
	}
	Broodwar->leaveGame();
}
void BasicAIModule::onFrame()
{
	tracer.onFrame();
}

void BasicAIModule::onUnitCreate(BWAPI::Unit* unit)
{
	tracer.onUnitCreate(unit);
}

void BasicAIModule::onUnitMorph(BWAPI::Unit* unit)
{
	tracer.onUnitMorph(unit);
}

void BasicAIModule::onUnitDestroy(BWAPI::Unit* unit)
{
	tracer.onUnitDestroy(unit);
}

void BasicAIModule::onUnitDiscover(BWAPI::Unit* unit)
{
	if (Broodwar->isReplay()) return;
}
void BasicAIModule::onUnitEvade(BWAPI::Unit* unit)
{
	if (Broodwar->isReplay()) return;
}

void BasicAIModule::onUnitRenegade(BWAPI::Unit* unit)
{
	if (Broodwar->isReplay()) return;
}

void BasicAIModule::onSendText(std::string text)
{	
	Broodwar->sendText("%s",text.c_str());
}

void BasicAIModule::stop_replays() const {
	const char * cmd = "copy /Y \"%STARCRAFT_LOCATION%\\bwapi-data\\stop_replays.ini\" \"%STARCRAFT_LOCATION%\\bwapi-data\\bwapi.ini\" && pause";
	int status = system(cmd);

	scai::log("copy returned %d", status);
}